#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import processutils as putils
from oslo_log import log as logging

from cinder import exception
from cinder.i18n import _
import cinder.privsep.targets.scst
from cinder.volume.targets import iscsi
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)


class SCSTAdm(iscsi.ISCSITarget):

    def __init__(self, *args, **kwargs):
        super(SCSTAdm, self).__init__(*args, **kwargs)
        self.volumes_dir = self.configuration.safe_get('volumes_dir')
        self.iscsi_target_prefix = self.configuration.safe_get(
            'target_prefix')
        self.target_name = self.configuration.safe_get('scst_target_iqn_name')
        self.target_driver = self.configuration.safe_get('scst_target_driver')
        self.chap_username = self.configuration.safe_get('chap_username')
        self.chap_password = self.configuration.safe_get('chap_password')
        self.initiator_iqn = None
        self.remove_initiator_iqn = None

    def scst_execute(self, *args):
        return cinder.privsep.targets.scst.run_scstadmin(*args)

    def validate_connector(self, connector):
        # iSCSI drivers require the initiator information
        if 'initiator' not in connector:
            err_msg = _('The volume driver requires the iSCSI initiator '
                        'name in the connector.')
            LOG.error(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)
        self.initiator_iqn = connector['initiator']

    def terminate_connection(self, volume, connector, **kwargs):
        self.remove_initiator_iqn = connector['initiator']

    def _get_target(self, iqn):
        (out, _err) = self.scst_execute('-list_target')
        if iqn in out:
            return self._target_attribute(iqn)
        return None

    def _target_attribute(self, iqn):
        (out, _err) = self.scst_execute('-list_tgt_attr', iqn,
                                        '-driver', self.target_driver)
        lines = out.split('\n')
        for line in lines:
            if "rel_tgt_id" in line:
                parsed = line.split()
                return parsed[1]

    def _get_group(self):
        scst_group = "%s%s" % (self.initiator_iqn, self.target_name)
        (out, _err) = self.scst_execute('-list_group')
        if scst_group in out:
            return out
        return None

    def _get_luns_info(self):
        scst_group = "%s%s" % (self.initiator_iqn, self.target_name)
        (out, _err) = self.scst_execute('-list_group', scst_group,
                                        '-driver', self.target_driver,
                                        '-target', self.target_name)

        first = "Assigned LUNs:"
        last = "Assigned Initiators:"
        start = out.index(first) + len(first)
        end = out.index(last, start)
        out = out[start:end]

        luns = []
        for line in out.strip().split("\n")[2:]:
            luns.append(int(line.strip().split(" ")[0]))
        luns = sorted(set(luns))
        return luns

    def _get_target_and_lun(self, context, volume):
        iscsi_target = 0
        if not self.target_name or not self._get_group():
            lun = 1
            return iscsi_target, lun

        luns = self._get_luns_info()
        if (not luns) or (luns[0] != 1):
            lun = 1
            return iscsi_target, lun
        else:
            for lun in luns:
                if (luns[-1] == lun) or (luns[lun - 1] + 1 != luns[lun]):
                    return iscsi_target, (lun + 1)

    def create_iscsi_target(self, name, vol_id, tid, lun, path,
                            chap_auth=None):
        scst_group = "%s%s" % (self.initiator_iqn, self.target_name)
        vol_name = path.split("/")[3]
        try:
            (out, _err) = self.scst_execute('-noprompt',
                                            '-set_drv_attr',
                                            self.target_driver,
                                            '-attributes',
                                            'enabled=1')
            LOG.debug('StdOut from set driver attribute: %s', out)
        except putils.ProcessExecutionError as e:
            LOG.error("Failed to set attribute for enable target driver %s", e)
            raise exception.ISCSITargetHelperCommandFailed(
                error_message="Failed to enable SCST Target driver.")

        if self._get_target(name) is None:
            try:
                (out, _err) = self.scst_execute('-add_target', name,
                                                '-driver', self.target_driver)
                LOG.debug("StdOut from scstadmin create target: %s", out)
            except putils.ProcessExecutionError as e:
                LOG.error("Failed to create iscsi target for volume "
                          "id:%(vol_id)s: %(e)s", {'vol_id': name, 'e': e})
                raise exception.ISCSITargetCreateFailed(volume_id=vol_name)
            try:
                (out, _err) = self.scst_execute('-enable_target', name,
                                                '-driver', self.target_driver)
                LOG.debug("StdOut from scstadmin enable target: %s", out)
            except putils.ProcessExecutionError as e:
                LOG.error("Failed to set 'enable' attribute for "
                          "SCST target %s", e)
                raise exception.ISCSITargetHelperCommandFailed(
                    error_message="Failed to enable SCST Target.")
            if chap_auth and self.target_name:
                try:
                    chap_string = self._iscsi_authentication('IncomingUser=',
                                                             *chap_auth)
                    (out, _err) = self.scst_execute('-noprompt',
                                                    '-set_tgt_attr', name,
                                                    '-driver',
                                                    self.target_driver,
                                                    '-attributes',
                                                    chap_string)
                    LOG.debug("StdOut from scstadmin set target attribute:"
                              " %s.", out)
                except putils.ProcessExecutionError:
                    msg = _("Failed to set attribute 'Incoming user' for "
                            "SCST target.")
                    LOG.exception(msg)
                    raise exception.ISCSITargetHelperCommandFailed(
                        error_message=msg)

        if self.target_name:
            if self._get_group() is None:
                try:
                    (out, _err) = self.scst_execute('-add_group', scst_group,
                                                    '-driver',
                                                    self.target_driver,
                                                    '-target', name)
                    LOG.debug("StdOut from scstadmin create group: %s", out)
                except putils.ProcessExecutionError as e:
                    LOG.error("Failed to create group to SCST target %s", e)
                    raise exception.ISCSITargetHelperCommandFailed(
                        error_message="Failed to create group to SCST target.")
            try:
                (out, _err) = self.scst_execute('-add_init',
                                                self.initiator_iqn,
                                                '-driver', self.target_driver,
                                                '-target', name,
                                                '-group', scst_group)
                LOG.debug("StdOut from scstadmin add initiator: %s", out)
            except putils.ProcessExecutionError as e:
                LOG.error("Failed to add initiator to group "
                          " for SCST target %s", e)
                raise exception.ISCSITargetHelperCommandFailed(
                    error_message="Failed to add Initiator to group for "
                                  "SCST target.")

        tid = self._get_target(name)
        if self.target_name is None:
            disk_id = "disk%s" % tid
        else:
            disk_id = "%s%s" % (lun, vol_id.split('-')[-1])

        try:
            self.scst_execute('-open_dev', disk_id,
                              '-handler', 'vdisk_fileio',
                              '-attributes', 'filename=%s' % path)
        except putils.ProcessExecutionError as e:
            LOG.error("Failed to add device to handler %s", e)
            raise exception.ISCSITargetHelperCommandFailed(
                error_message="Failed to add device to SCST handler.")

        try:
            if self.target_name:
                self.scst_execute('-add_lun', lun,
                                  '-driver', self.target_driver,
                                  '-target', name,
                                  '-device', disk_id,
                                  '-group', scst_group)
            else:
                self.scst_execute('-add_lun', lun,
                                  '-driver', self.target_driver,
                                  '-target', name,
                                  '-device', disk_id)
        except putils.ProcessExecutionError as e:
            LOG.error("Failed to add lun to SCST target "
                      "id:%(vol_id)s: %(e)s", {'vol_id': name, 'e': e})
            raise exception.ISCSITargetHelperCommandFailed(
                error_message="Failed to add LUN to SCST Target for "
                              "volume " + vol_name)

            # SCST uses /etc/scst.conf as the default configuration when it
            # starts
        try:
            self.scst_execute('-write_config', '/etc/scst.conf')
        except putils.ProcessExecutionError:
            LOG.error("Failed to write in /etc/scst.conf.")
            raise exception.ISCSITargetHelperCommandFailed(
                error_message="Failed to write in /etc/scst.conf.")

        return tid

    def _iscsi_location(self, ip, target, iqn, lun=None):
        return "%s:%s,%s %s %s" % (volume_utils.sanitize_host(ip),
                                   self.configuration.target_port,
                                   target, iqn, lun)

    def _get_iscsi_name(self, volume):
        if self.target_name is None:
            return "%s%s" % (self.configuration.target_prefix,
                             volume['name'])
        else:
            return self.target_name

    def _get_iscsi_target(self, context, vol_id):
        # FIXME(jdg): Need to implement abc method
        pass

    def _get_target_chap_auth(self, context, volume):
        # FIXME(jdg): Need to implement abc method

        iscsi_name = self._get_iscsi_name(volume)

        if self._get_target(iscsi_name) is None:
            return None
        (out, _err) = self.scst_execute('-list_tgt_attr', iscsi_name,
                                        '-driver', self.target_driver)
        first = "KEY"
        last = "Dynamic attributes"
        start = out.index(first) + len(first)
        end = out.index(last, start)
        out = out[start:end]
        out = out.split("\n")[2]
        if "IncomingUser" in out:
            out = out.split(" ")
            out = [a for a in out if a != ""]
            return (out[1], out[2])
        else:
            return None

    def ensure_export(self, context, volume, volume_path):
        iscsi_target, lun = self._get_target_and_lun(context, volume)
        iscsi_name = self._get_iscsi_name(volume)

        if self.chap_username and self.chap_password:
            chap_auth = (self.chap_username, self.chap_password)
        else:
            chap_auth = self._get_target_chap_auth(context, volume)

        self.create_iscsi_target(iscsi_name, volume['id'], iscsi_target,
                                 lun, volume_path, chap_auth)

    def create_export(self, context, volume, volume_path):
        """Creates an export for a logical volume."""
        iscsi_target, lun = self._get_target_and_lun(context, volume)
        iscsi_name = self._get_iscsi_name(volume)

        if self.chap_username and self.chap_password:
            chap_auth = (self.chap_username, self.chap_password)
        else:
            chap_auth = self._get_target_chap_auth(context, volume)
            if not chap_auth:
                chap_auth = (volume_utils.generate_username(),
                             volume_utils.generate_password())
        tid = self.create_iscsi_target(iscsi_name, volume['id'], iscsi_target,
                                       lun, volume_path, chap_auth)

        data = {}
        data['location'] = self._iscsi_location(
            self.configuration.target_ip_address, tid, iscsi_name, lun)
        LOG.debug('Set provider_location to: %s', data['location'])
        data['auth'] = self._iscsi_authentication(
            'CHAP', *chap_auth)
        return data

    def remove_export(self, context, volume):
        try:
            location = volume['provider_location'].split(' ')
            iqn = location[1]
            iscsi_target = self._get_target(iqn)
            self.show_target(iscsi_target, iqn)

        except Exception:
            LOG.error("Skipping remove_export. No iscsi_target is "
                      "presently exported for volume: %s", volume['id'])
            return
        vol = self.db.volume_get(context, volume['id'])
        lun = "".join(vol['provider_location'].split(" ")[-1:])

        self.remove_iscsi_target(iscsi_target, lun,
                                 volume['id'], volume['name'])

    def remove_iscsi_target(self, tid, lun, vol_id, vol_name, **kwargs):
        disk_id = "%s%s" % (lun, vol_id.split('-')[-1])
        vol_uuid_file = vol_name
        if self.target_name is None:
            iqn = '%s%s' % (self.iscsi_target_prefix, vol_uuid_file)
        else:
            iqn = self.target_name

        if self.target_name is None:
            try:
                self.scst_execute('-noprompt',
                                  '-rem_target', iqn,
                                  '-driver', 'iscsi')
            except putils.ProcessExecutionError as e:
                LOG.error("Failed to remove iscsi target for volume "
                          "id:%(vol_id)s: %(e)s", {'vol_id': vol_id, 'e': e})
                raise exception.ISCSITargetRemoveFailed(volume_id=vol_id)
            try:
                self.scst_execute('-noprompt',
                                  '-close_dev', "disk%s" % tid,
                                  '-handler', 'vdisk_fileio')
            except putils.ProcessExecutionError as e:
                LOG.error("Failed to close disk device %s", e)
                raise exception.ISCSITargetHelperCommandFailed(
                    error_message="Failed to close disk device for "
                                  "SCST handler.")

            if self._get_target(iqn):
                try:
                    self.scst_execute('-noprompt',
                                      '-rem_target', iqn,
                                      '-driver', self.target_driver)
                except putils.ProcessExecutionError as e:
                    LOG.error("Failed to remove iscsi target for "
                              "volume id:%(vol_id)s: %(e)s",
                              {'vol_id': vol_id, 'e': e})
                    raise exception.ISCSITargetRemoveFailed(volume_id=vol_id)
        else:
            if not int(lun) in self._get_luns_info():
                raise exception.ISCSITargetRemoveFailed(volume_id=vol_id)
            try:
                scst_group = "%s%s" % (self.remove_initiator_iqn,
                                       self.target_name)
                self.scst_execute('-noprompt', '-rem_lun', lun,
                                  '-driver', self.target_driver,
                                  '-target', iqn, '-group',
                                  scst_group)
            except putils.ProcessExecutionError as e:
                LOG.error("Failed to remove LUN %s", e)
                raise exception.ISCSITargetHelperCommandFailed(
                    error_message="Failed to remove LUN for SCST Target.")

            try:
                self.scst_execute('-noprompt',
                                  '-close_dev', disk_id,
                                  '-handler', 'vdisk_fileio')
            except putils.ProcessExecutionError as e:
                LOG.error("Failed to close disk device %s", e)
                raise exception.ISCSITargetHelperCommandFailed(
                    error_message="Failed to close disk device for "
                                  "SCST handler.")

        self.scst_execute('-write_config', '/etc/scst.conf')

    def show_target(self, tid, iqn):
        if iqn is None:
            raise exception.InvalidParameterValue(
                err=_('valid iqn needed for show_target'))

        tid = self._get_target(iqn)
        if tid is None:
            raise exception.ISCSITargetHelperCommandFailed(
                error_message="Target not found")



#  Copyright (c) 2016 Stratoscale, Ltd.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import oslo_messaging as messaging

from cinder.api import common
from cinder.api import microversions as mv
from cinder import exception
from cinder.i18n import _


def get_manageable_resources(req, is_detail, function_get_manageable,
                             view_builder):
    context = req.environ['cinder.context']
    params = req.params.copy()
    cluster_name, host = common.get_cluster_host(
        req, params, mv.MANAGE_EXISTING_CLUSTER)
    marker, limit, offset = common.get_pagination_params(params)
    sort_keys, sort_dirs = common.get_sort_params(params,
                                                  default_key='reference')

    # These parameters are generally validated at the DB layer, but in this
    # case sorting is not done by the DB
    valid_sort_keys = ('reference', 'size')
    invalid_keys = [key for key in sort_keys if key not in valid_sort_keys]
    if invalid_keys:
        msg = _("Invalid sort keys passed: %s") % ', '.join(invalid_keys)
        raise exception.InvalidParameterValue(err=msg)
    valid_sort_dirs = ('asc', 'desc')
    invalid_dirs = [d for d in sort_dirs if d not in valid_sort_dirs]
    if invalid_dirs:
        msg = _("Invalid sort dirs passed: %s") % ', '.join(invalid_dirs)
        raise exception.InvalidParameterValue(err=msg)

    try:
        resources = function_get_manageable(context, host, cluster_name,
                                            marker=marker, limit=limit,
                                            offset=offset, sort_keys=sort_keys,
                                            sort_dirs=sort_dirs)
    except messaging.RemoteError as err:
        if err.exc_type == "InvalidInput":
            raise exception.InvalidInput(err.value)
        raise

    resource_count = len(resources)

    if is_detail:
        resources = view_builder.detail_list(req, resources, resource_count)
    else:
        resources = view_builder.summary_list(req, resources, resource_count)
    return resources



#    (c) Copyright 2014 Brocade Communications Systems Inc.
#    All Rights Reserved.
#
#    Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import re

from oslo_log import log

LOG = log.getLogger(__name__)


def get_friendly_zone_name(zoning_policy, initiator, target,
                           host_name, storage_system, zone_name_prefix,
                           supported_chars):
    """Utility function implementation of _get_friendly_zone_name.

    Get friendly zone name is used to form the zone name
    based on the details provided by the caller

    :param zoning_policy: determines the zoning policy is either
                          initiator-target or initiator
    :param initiator: initiator WWN
    :param target: target WWN
    :param host_name: Host name returned from Volume Driver
    :param storage_system: Storage name returned from Volume Driver
    :param zone_name_prefix: user defined zone prefix configured
                             in cinder.conf
    :param supported_chars: Supported character set of FC switch vendor.
                            Example: `abc123_-$`. These are defined in
                            the FC zone drivers.
    """
    if host_name is None:
        host_name = ''
    if storage_system is None:
        storage_system = ''
    if zoning_policy == 'initiator-target':
        host_name = host_name[:14]
        storage_system = storage_system[:14]
        if len(host_name) > 0 and len(storage_system) > 0:
            zone_name = (host_name + "_"
                         + initiator.replace(':', '') + "_"
                         + storage_system + "_"
                         + target.replace(':', ''))
        else:
            zone_name = (zone_name_prefix
                         + initiator.replace(':', '')
                         + target.replace(':', ''))
            LOG.info("Zone name created using prefix because either "
                     "host name or storage system is none.")
    else:
        host_name = host_name[:47]
        if len(host_name) > 0:
            zone_name = (host_name + "_"
                         + initiator.replace(':', ''))
        else:
            zone_name = (zone_name_prefix
                         + initiator.replace(':', ''))
            LOG.info("Zone name created using prefix because host "
                     "name is none.")

    LOG.info("Friendly zone name after forming: %(zonename)s",
             {'zonename': zone_name})
    zone_name = re.sub('[^%s]' % supported_chars, '', zone_name)
    return zone_name



# Copyright (c) - 2014, Clinton Knight  All rights reserved.
# Copyright (c) - 2015, Alex Meade.  All Rights Reserved.
# Copyright (c) - 2015, Rushil Chugh.  All Rights Reserved.
# Copyright (c) - 2015, Tom Barron.  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from cinder.volume import configuration as conf
import cinder.volume.drivers.netapp.options as na_opts


ISCSI_FAKE_LUN_ID = 1

ISCSI_FAKE_IQN = 'iqn.1993-08.org.debian:01:10'
ISCSI_FAKE_IQN2 = 'iqn.1993-08.org.debian:01:11'

ISCSI_FAKE_ADDRESS_IPV4 = '10.63.165.216'
ISCSI_FAKE_ADDRESS2_IPV4 = '10.63.165.217'
ISCSI_FAKE_ADDRESS_IPV6 = 'fe80::72a4:a152:aad9:30d9'

ISCSI_FAKE_PORT = '2232'

ISCSI_FAKE_VOLUME = {'id': 'fake_id'}

ISCSI_FAKE_TARGET = {}
ISCSI_FAKE_TARGET['address'] = ISCSI_FAKE_ADDRESS_IPV4
ISCSI_FAKE_TARGET['port'] = ISCSI_FAKE_PORT

ISCSI_FAKE_VOLUME = {'id': 'fake_id', 'provider_auth': 'None stack password'}
ISCSI_FAKE_VOLUME_NO_AUTH = {'id': 'fake_id', 'provider_auth': ''}

ISCSI_MP_TARGET_INFO_DICT = {'target_discovered': False,
                             'target_portal': '10.63.165.216:2232',
                             'target_portals': ['10.63.165.216:2232',
                                                '10.63.165.217:2232'],
                             'target_iqn': ISCSI_FAKE_IQN,
                             'target_iqns': [ISCSI_FAKE_IQN, ISCSI_FAKE_IQN2],
                             'target_lun': ISCSI_FAKE_LUN_ID,
                             'target_luns': [ISCSI_FAKE_LUN_ID] * 2,
                             'volume_id': ISCSI_FAKE_VOLUME['id'],
                             'auth_method': 'None', 'auth_username': 'stack',
                             'auth_password': 'password'}

FC_ISCSI_TARGET_INFO_DICT = {'target_discovered': False,
                             'target_portal': '10.63.165.216:2232',
                             'target_iqn': ISCSI_FAKE_IQN,
                             'target_lun': ISCSI_FAKE_LUN_ID,
                             'volume_id': ISCSI_FAKE_VOLUME['id'],
                             'auth_method': 'None', 'auth_username': 'stack',
                             'auth_password': 'password'}

FC_ISCSI_TARGET_INFO_DICT_IPV6 = {'target_discovered': False,
                                  'target_portal':
                                      '[fe80::72a4:a152:aad9:30d9]:2232',
                                  'target_iqn': ISCSI_FAKE_IQN,
                                  'target_lun': ISCSI_FAKE_LUN_ID,
                                  'volume_id': ISCSI_FAKE_VOLUME['id']}

VOLUME_NAME = 'fake_volume_name'
VOLUME_ID = 'fake_volume_id'
VOLUME_TYPE_ID = 'fake_volume_type_id'

VOLUME = {
    'name': VOLUME_NAME,
    'size': 42,
    'id': VOLUME_ID,
    'host': 'fake_host@fake_backend#fake_pool',
    'volume_type_id': VOLUME_TYPE_ID,
}

SNAPSHOT_NAME = 'fake_snapshot_name'
SNAPSHOT_ID = 'fake_snapshot_id'

SNAPSHOT = {
    'name': SNAPSHOT_NAME,
    'id': SNAPSHOT_ID,
    'volume_id': VOLUME_ID,
    'volume_name': VOLUME_NAME,
    'volume_size': 42,
}

QOS_SPECS = {}

EXTRA_SPECS = {}

MAX_THROUGHPUT = '21734278B/s'
QOS_POLICY_GROUP_NAME = 'fake_qos_policy_group_name'
LEGACY_EXTRA_SPECS = {'netapp:qos_policy_group': QOS_POLICY_GROUP_NAME}

LEGACY_QOS = {
    'policy_name': QOS_POLICY_GROUP_NAME,
}

QOS_POLICY_GROUP_SPEC = {
    'max_throughput': MAX_THROUGHPUT,
    'policy_name': 'openstack-%s' % VOLUME_ID,
}

QOS_POLICY_GROUP_INFO_NONE = {'legacy': None, 'spec': None}

QOS_POLICY_GROUP_INFO = {'legacy': None, 'spec': QOS_POLICY_GROUP_SPEC}

LEGACY_QOS_POLICY_GROUP_INFO = {
    'legacy': LEGACY_QOS,
    'spec': None,
}

INVALID_QOS_POLICY_GROUP_INFO = {
    'legacy': LEGACY_QOS,
    'spec': QOS_POLICY_GROUP_SPEC,
}

QOS_SPECS_ID = 'fake_qos_specs_id'
QOS_SPEC = {'maxBPS': 21734278}
OUTER_BACKEND_QOS_SPEC = {
    'id': QOS_SPECS_ID,
    'specs': QOS_SPEC,
    'consumer': 'back-end',
}
OUTER_FRONTEND_QOS_SPEC = {
    'id': QOS_SPECS_ID,
    'specs': QOS_SPEC,
    'consumer': 'front-end',
}
OUTER_BOTH_QOS_SPEC = {
    'id': QOS_SPECS_ID,
    'specs': QOS_SPEC,
    'consumer': 'both',
}
VOLUME_TYPE = {'id': VOLUME_TYPE_ID, 'qos_specs_id': QOS_SPECS_ID}


def create_configuration():
    config = conf.Configuration(None)
    config.append_config_values(na_opts.netapp_connection_opts)
    config.append_config_values(na_opts.netapp_transport_opts)
    config.append_config_values(na_opts.netapp_basicauth_opts)
    config.append_config_values(na_opts.netapp_provisioning_opts)
    return config


def create_configuration_cmode():
    config = create_configuration()
    config.append_config_values(na_opts.netapp_cluster_opts)
    return config



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Cinder OS API WSGI application."""

import sys
import warnings
warnings.simplefilter('once', DeprecationWarning)

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import wsgi

from cinder import objects  # noqa
from cinder import i18n  # noqa
i18n.enable_lazy()
# Need to register global_opts
from cinder.common import config
from cinder.common import constants
from cinder import coordination
from cinder import rpc
from cinder import service
from cinder import version

CONF = cfg.CONF


def initialize_application():
    objects.register_all()
    CONF(sys.argv[1:], project='cinder',
         version=version.version_string())
    logging.setup(CONF, "cinder")
    config.set_middleware_defaults()

    coordination.COORDINATOR.start()

    rpc.init(CONF)
    service.setup_profiler(constants.API_BINARY, CONF.host)

    return wsgi.Loader(CONF).load_app(name='osapi_volume')



#  Copyright (c) 2016 IBM Corporation
#  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from cinder import exception
from cinder.i18n import _

BLOCKS_PER_GIGABYTE = 2097152
XIV_LOG_PREFIX = "[IBM XIV STORAGE]:"
XIV_CONNECTION_TYPE_ISCSI = 'iscsi'
XIV_CONNECTION_TYPE_FC = 'fibre_channel'
XIV_CONNECTION_TYPE_FC_ECKD = 'fibre_channel_eckd'
CHAP_NONE = 'disabled'
CHAP_ENABLED = 'enabled'
STORAGE_DRIVER_XIV = 'xiv'
STORAGE_DRIVER_DS8K = 'ds8k'


CONF_KEYS = {
    'driver': "volume_driver",
    'proxy': "proxy",
    'user': "san_login",
    'password': "san_password",
    'storage_pool': "san_clustername",
    'address': "san_ip",
    'driver_version': "ibm_storage_driver_version",
    'volume_api_class': "volume_api_class",
    'volume_backend': "volume_backend_name",
    'connection_type': "connection_type",
    'management_ips': "management_ips",
    'chap': 'chap',
    'system_id': 'system_id',
    'replication_device': 'replication_device'
}
CONF_BACKEND_KEYS = {
    'user': "san_login",
    'password': "san_password",
    'storage_pool': "san_clustername",
    'address': "san_ip",
    'volume_backend': "volume_backend_name",
    'connection_type': "connection_type",
    'management_ips': "management_ips",
}
FLAG_KEYS = {
    'user': "user",
    'password': "password",
    'storage_pool': "vol_pool",
    'address': "address",
    'connection_type': "connection_type",
    'bypass_connection_check': "XIV_BYPASS_CONNECTION_CHECK",
    'management_ips': "management_ips"
}
METADATA_KEYS = {
    'ibm_storage_version': 'openstack_ibm_storage_driver_version',
    'openstack_version': 'openstack_version',
    'pool_host_key': 'openstack_compute_node_%(hostname)s',
    'pool_volume_os': 'openstack_volume_os',
    'pool_volume_hostname': 'openstack_volume_hostname'
}


def get_host_or_create_from_iqn(connector, connection=None):
    """Get host name.

    Return the hostname if existing at the connector (nova-compute info)
    If not, generate one from the IQN or HBA
    """
    if connection is None and connector.get('host', None):
        return connector['host']

    if connection != XIV_CONNECTION_TYPE_FC and 'initiator' in connector:
        try:
            initiator = connector['initiator']
            iqn_suffix = initiator.split('.')[-1].replace(":", "_")
        except Exception:
            if connector.get('initiator', 'None'):
                raise exception.VolumeDriverException(message=(
                    _("Initiator format: %(iqn)s")) %
                    {'iqn': connector.get('initiator', 'None')})
            else:
                raise exception.VolumeDriverException(
                    message=_("Initiator is missing from connector object"))
        return "nova-compute-%s" % iqn_suffix

    if connection != XIV_CONNECTION_TYPE_ISCSI and len(
        connector.get('wwpns', [])
    ) > 0:
        return "nova-compute-%s" % connector['wwpns'][0].replace(":", "_")

    raise exception.VolumeDriverException(
        message=_("Compute host missing either iSCSI initiator or FC wwpns"))


def gigabytes_to_blocks(gigabytes):
    return int(BLOCKS_PER_GIGABYTE * float(gigabytes))


def get_online_iscsi_ports(ibm_storage_cli):
    """Returns online iscsi ports."""
    iscsi_ports = [
        {
            'ip': p.get('address'),
            # ipinterface_list returns ports field in Gen3, and
            # port field in BlueRidge
            'port': p.get('ports', p.get('port')),
            'module': p.get('module')
        } for p in ibm_storage_cli.cmd.ipinterface_list()
        if p.type == 'iSCSI']

    iscsi_connected_ports = [
        {
            'port': p.index,
            'module': p.get('module_id')
        } for p in ibm_storage_cli.cmd.ipinterface_list_ports()
        if p.is_link_up == 'yes' and p.role == 'iSCSI']

    to_return = []
    for ip in iscsi_ports:
        if len([
            p for p in iscsi_connected_ports
            if (p.get('port') == ip.get('port') and
                p.get('module') == ip.get('module'))
        ]) > 0:
            to_return += [ip.get('ip')]

    return to_return



# Copyright (C) 2020, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""REST interface fibre channel module for Hitachi HBSD Driver."""

from oslo_log import log as logging

from cinder.volume.drivers.hitachi import hbsd_rest as rest
from cinder.volume.drivers.hitachi import hbsd_utils as utils
from cinder.zonemanager import utils as fczm_utils

_FC_HMO_DISABLE_IO = 91

LOG = logging.getLogger(__name__)
MSG = utils.HBSDMsg


class HBSDRESTFC(rest.HBSDREST):
    """REST interface fibre channel class for Hitachi HBSD Driver."""

    def __init__(self, conf, storage_protocol, db):
        """Initialize instance variables."""
        super(HBSDRESTFC, self).__init__(conf, storage_protocol, db)
        self._lookup_service = fczm_utils.create_lookup_service()

    def connect_storage(self):
        """Prepare for using the storage."""
        target_ports = self.conf.hitachi_target_ports
        compute_target_ports = self.conf.hitachi_compute_target_ports
        available_ports = []
        available_compute_ports = []

        super(HBSDRESTFC, self).connect_storage()
        # The port attributes must contain TAR.
        params = {'portAttributes': 'TAR'}
        port_list = self.client.get_ports(params=params)
        for port in set(target_ports + compute_target_ports):
            if port not in [port_data['portId'] for port_data in port_list]:
                utils.output_log(MSG.INVALID_PORT, port=port,
                                 additional_info='portAttributes: not TAR')
        for port_data in port_list:
            port = port_data['portId']
            if port not in set(target_ports + compute_target_ports):
                continue
            secure_fc_port = True
            if (port_data['portType'] not in ['FIBRE', 'FCoE'] or
                    not port_data['lunSecuritySetting']):
                secure_fc_port = False
            if not secure_fc_port:
                utils.output_log(
                    MSG.INVALID_PORT, port=port,
                    additional_info='portType: %s, lunSecuritySetting: %s, '
                    'fabricMode: %s, portConnection: %s' %
                    (port_data['portType'],
                     port_data.get('lunSecuritySetting'),
                     port_data.get('fabricMode'),
                     port_data.get('portConnection')))
            if not secure_fc_port:
                continue
            wwn = port_data.get('wwn')
            if target_ports and port in target_ports:
                available_ports.append(port)
                self.storage_info['wwns'][port] = wwn
            if compute_target_ports and port in compute_target_ports:
                available_compute_ports.append(port)
                self.storage_info['wwns'][port] = wwn

        if target_ports:
            for port in target_ports:
                if port in available_ports:
                    self.storage_info['controller_ports'].append(port)
        if compute_target_ports:
            for port in compute_target_ports:
                if port in available_compute_ports:
                    self.storage_info['compute_ports'].append(port)

        self.check_ports_info()
        utils.output_log(MSG.SET_CONFIG_VALUE, object='port-wwn list',
                         value=self.storage_info['wwns'])

    def create_target_to_storage(self, port, connector, hba_ids):
        """Create a host group on the specified port."""
        wwpns = self.get_hba_ids_from_connector(connector)
        target_name = '%(prefix)s-%(wwpns)s' % {
            'prefix': utils.DRIVER_PREFIX,
            'wwpns': min(wwpns),
        }
        try:
            body = {'portId': port,
                    'hostGroupName': target_name}
            gid = self.client.add_host_grp(body, no_log=True)
        except Exception:
            params = {'portId': port}
            host_grp_list = self.client.get_host_grps(params)
            for host_grp_data in host_grp_list:
                if host_grp_data['hostGroupName'] == target_name:
                    return target_name, host_grp_data['hostGroupNumber']
            raise
        return target_name, gid

    def set_hba_ids(self, port, gid, hba_ids):
        """Connect all specified HBAs with the specified port."""
        registered_wwns = []
        for wwn in hba_ids:
            try:
                self.client.add_hba_wwn(port, gid, wwn, no_log=True)
                registered_wwns.append(wwn)
            except utils.HBSDError:
                utils.output_log(MSG.ADD_HBA_WWN_FAILED, port=port, gid=gid,
                                 wwn=wwn)
        if not registered_wwns:
            msg = utils.output_log(MSG.NO_HBA_WWN_ADDED_TO_HOST_GRP, port=port,
                                   gid=gid)
            raise utils.HBSDError(msg)

    def set_target_mode(self, port, gid):
        """Configure the host group to meet the environment."""
        body = {'hostMode': 'LINUX/IRIX',
                'hostModeOptions': [_FC_HMO_DISABLE_IO]}
        self.client.modify_host_grp(port, gid, body, ignore_all_errors=True)

    def _get_hwwns_in_hostgroup(self, port, gid, wwpns):
        """Return WWN registered with the host group."""
        hwwns_in_hostgroup = []
        for hba_wwn in self.client.get_hba_wwns(port, gid):
            hwwn = hba_wwn['hostWwn']
            if hwwn in wwpns:
                hwwns_in_hostgroup.append(hwwn)
        return hwwns_in_hostgroup

    def _set_target_info(self, targets, host_grps, wwpns):
        """Set the information of the host group having the specified WWN."""
        for host_grp in host_grps:
            port = host_grp['portId']
            gid = host_grp['hostGroupNumber']
            hwwns_in_hostgroup = self._get_hwwns_in_hostgroup(port, gid, wwpns)
            if hwwns_in_hostgroup:
                targets['info'][port] = True
                targets['list'].append((port, gid))
                LOG.debug(
                    'Found wwpns in host group. (port: %(port)s, '
                    'gid: %(gid)s, wwpns: %(wwpns)s)',
                    {'port': port, 'gid': gid, 'wwpns': hwwns_in_hostgroup})
                return True
        return False

    def _get_hwwns_in_hostgroup_by_name(self, port, host_group_name, wwpns):
        """Return WWN registered with the host group of the specified name."""
        hba_wwns = self.client.get_hba_wwns_by_name(port, host_group_name)
        return [hba_wwn for hba_wwn in hba_wwns if hba_wwn['hostWwn'] in wwpns]

    def _set_target_info_by_names(self, targets, port, target_names, wwpns):
        """Set the information of the host group having the specified name and

        the specified WWN.
        """
        for target_name in target_names:
            hwwns_in_hostgroup = self._get_hwwns_in_hostgroup_by_name(
                port, target_name, wwpns)
            if hwwns_in_hostgroup:
                gid = hwwns_in_hostgroup[0]['hostGroupNumber']
                targets['info'][port] = True
                targets['list'].append((port, gid))
                LOG.debug(
                    'Found wwpns in host group. (port: %(port)s, '
                    'gid: %(gid)s, wwpns: %(wwpns)s)',
                    {'port': port, 'gid': gid, 'wwpns':
                     [hwwn['hostWwn'] for hwwn in hwwns_in_hostgroup]})
                return True
        return False

    def find_targets_from_storage(
            self, targets, connector, target_ports):
        """Find mapped ports, memorize them and return unmapped port count."""
        wwpns = self.get_hba_ids_from_connector(connector)
        target_names = [
            '%(prefix)s-%(wwpns)s' % {
                'prefix': utils.DRIVER_PREFIX,
                'wwpns': min(wwpns),
            }
        ]
        if 'ip' in connector:
            target_names.append(
                '%(prefix)s-%(ip)s' % {
                    'prefix': utils.DRIVER_PREFIX,
                    'ip': connector['ip'],
                }
            )
        not_found_count = 0
        for port in target_ports:
            targets['info'][port] = False
            if self._set_target_info_by_names(
                    targets, port, target_names, wwpns):
                continue
            host_grps = self.client.get_host_grps({'portId': port})
            if self._set_target_info(
                targets, [hg for hg in host_grps if hg['hostGroupName'] not in
                          target_names], wwpns):
                pass
            else:
                not_found_count += 1
        return not_found_count

    def initialize_connection(self, volume, connector):
        """Initialize connection between the server and the volume."""
        conn_info = super(HBSDRESTFC, self).initialize_connection(
            volume, connector)
        if self.conf.hitachi_zoning_request:
            init_targ_map = utils.build_initiator_target_map(
                connector, conn_info['data']['target_wwn'],
                self._lookup_service)
            if init_targ_map:
                conn_info['data']['initiator_target_map'] = init_targ_map
            fczm_utils.add_fc_zone(conn_info)
        return conn_info

    def terminate_connection(self, volume, connector):
        """Terminate connection between the server and the volume."""
        conn_info = super(HBSDRESTFC, self).terminate_connection(
            volume, connector)
        if self.conf.hitachi_zoning_request:
            if conn_info and conn_info['data']['target_wwn']:
                init_targ_map = utils.build_initiator_target_map(
                    connector, conn_info['data']['target_wwn'],
                    self._lookup_service)
                if init_targ_map:
                    conn_info['data']['initiator_target_map'] = init_targ_map
            fczm_utils.remove_fc_zone(conn_info)
        return conn_info

    def _get_wwpns(self, port, hostgroup):
        """Get WWPN from a port and the host group."""
        wwpns = []
        hba_wwns = self.client.get_hba_wwns_by_name(port, hostgroup)
        for hba_wwn in hba_wwns:
            wwpns.append(hba_wwn['hostWwn'])
        return wwpns

    def set_terminate_target(self, fake_connector, port_hostgroup_map):
        """Set necessary information in connector in terminate."""
        wwpns = set()
        for port, hostgroups in port_hostgroup_map.items():
            for hostgroup in hostgroups:
                wwpns.update(self._get_wwpns(port, hostgroup))
        fake_connector['wwpns'] = list(wwpns)



# Copyright (c) 2014 ProphetStor, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg

from cinder.volume import configuration

DPL_OPTS = [
    cfg.StrOpt('dpl_pool',
               default='',
               help='DPL pool uuid in which DPL volumes are stored.'),
    cfg.PortOpt('dpl_port',
                default=8357,
                help='DPL port number.'),
]

CONF = cfg.CONF
CONF.register_opts(DPL_OPTS, group=configuration.SHARED_CONF_GROUP)



# Copyright 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cinder.common import constants
from cinder import service
from cinder.tests.unit import test


class SetupProfilerTestCase(test.TestCase):
    def setUp(self):
        super(SetupProfilerTestCase, self).setUp()
        service.osprofiler_initializer = mock.MagicMock()
        service.profiler = mock.MagicMock()
        service.profiler_opts = mock.MagicMock()
        service.osprofiler_initializer.init_from_conf = mock.MagicMock()

    def test_profiler_not_present(self):
        service.profiler = None
        service.LOG.debug = mock.MagicMock()
        service.setup_profiler(constants.VOLUME_BINARY, "localhost")
        service.LOG.debug.assert_called_once_with("osprofiler is not present")

    @mock.patch("cinder.service.context")
    def test_profiler_enabled(self, context):
        service.CONF.profiler.enabled = True
        return_value = {"Meaning Of Life": 42}
        context.get_admin_context().to_dict.return_value = return_value
        service.setup_profiler(constants.VOLUME_BINARY, "localhost")
        service.osprofiler_initializer.init_from_conf.assert_called_once_with(
            conf=service.CONF,
            context=return_value,
            project="cinder",
            service=constants.VOLUME_BINARY,
            host="localhost")

    def test_profiler_disabled(self):
        service.CONF.profiler.enabled = False
        service.setup_profiler(constants.VOLUME_BINARY, "localhost")
        service.osprofiler_initializer.init_from_conf.assert_not_called()



# Copyright 2018 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy

from cinder.api.validation import parameter_types


enable_and_disable = {
    'type': 'object',
    'properties': {
        'binary': {'type': 'string', 'minLength': 1, 'maxLength': 255},
        'host': parameter_types.hostname,
        'cluster': parameter_types.nullable_string,
        'service': {'type': 'string', 'minLength': 1, 'maxLength': 255},
    },
    'anyOf': [
        {'required': ['binary']},
        {'required': ['service']}
    ],
    'additionalProperties': False,
}


disable_log_reason = copy.deepcopy(enable_and_disable)
disable_log_reason['properties'][
    'disabled_reason'] = {'type': 'string', 'minLength': 1, 'maxLength': 255,
                          'format': 'disabled_reason'}


set_log = {
    'type': 'object',
    'properties': {
        'binary': parameter_types.binary,
        'server': parameter_types.nullable_string,
        'prefix': parameter_types.nullable_string,
        'level': {'type': ['string', 'null'], 'format': 'level'}
    },
    'additionalProperties': False,
}


get_log = {
    'type': 'object',
    'properties': {
        'binary': parameter_types.binary,
        'server': parameter_types.nullable_string,
        'prefix': parameter_types.nullable_string,
    },
    'additionalProperties': False,
}


freeze_and_thaw = {
    'type': 'object',
    'properties': {
        'cluster': parameter_types.nullable_string,
        'host': parameter_types.hostname,
    },
    'additionalProperties': False,
}


failover_host = {
    'type': 'object',
    'properties': {
        'host': parameter_types.hostname,
        'backend_id': parameter_types.nullable_string,
        'cluster': parameter_types.nullable_string,
    },
    'additionalProperties': False,
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from os_brick import initiator
from os_brick.initiator import connector
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import units
import requests

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import interface
from cinder import utils
from cinder.volume import driver
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)


@interface.volumedriver
class SPDKDriver(driver.VolumeDriver):
    """Executes commands relating to Volumes."""

    VERSION = '1.0.0'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = "Mellanox_CI"

    def __init__(self, *args, **kwargs):
        # Parent sets db, host, _execute and base config
        super(SPDKDriver, self).__init__(*args, **kwargs)

        self.lvs = []
        self.ctxt = context.get_admin_context()

        target_driver = (
            self.target_mapping[self.configuration.safe_get('target_helper')])

        LOG.debug('SPDK attempting to initialize LVM driver with the '
                  'following target_driver: %s',
                  target_driver)

        self.target_driver = importutils.import_object(
            target_driver,
            configuration=self.configuration,
            db=self.db,
            executor=self._execute)

    @staticmethod
    def get_driver_options():
        return []

    def _rpc_call(self, method, params=None):
        payload = {}
        payload['jsonrpc'] = '2.0'
        payload['id'] = 1
        payload['method'] = method
        if params is not None:
            payload['params'] = params

        req = requests.post(self.url,
                            data=json.dumps(payload),
                            auth=(self.configuration.spdk_rpc_username,
                                  self.configuration.spdk_rpc_password),
                            verify=self.configuration.driver_ssl_cert_verify,
                            timeout=30)

        if not req.ok:
            raise exception.VolumeBackendAPIException(
                data=_('SPDK target responded with error: %s') % req.text)

        return req.json()['result']

    def _update_volume_stats(self):
        """Retrieve stats info from volume group."""

        LOG.debug('SPDK Updating volume stats')
        status = {'volume_backend_name': 'SPDK',
                  'vendor_name': 'Open Source',
                  'driver_version': self.VERSION,
                  'storage_protocol': 'NVMe-oF'}
        pools_status = []
        self.lvs = []

        output = self._rpc_call('bdev_lvol_get_lvstores')
        if output:
            for lvs in output:
                pool = {}
                lvs_entry = {}
                free_size = (lvs['free_clusters']
                             * lvs['cluster_size']
                             / units.Gi)
                total_size = (lvs['total_data_clusters']
                              * lvs['cluster_size']
                              / units.Gi)
                pool["volume_backend_name"] = 'SPDK'
                pool["vendor_name"] = 'Open Source'
                pool["driver_version"] = self.VERSION
                pool["storage_protocol"] = 'NVMe-oF'
                pool["total_capacity_gb"] = total_size
                pool["free_capacity_gb"] = free_size
                pool["pool_name"] = lvs['name']
                pools_status.append(pool)

                lvs_entry['name'] = lvs['name']
                lvs_entry['uuid'] = lvs['uuid']
                lvs_entry['free_size'] = free_size
                lvs_entry['total_size'] = total_size
                self.lvs.append(lvs_entry)

        status['pools'] = pools_status
        self._stats = status

        for lvs in self.lvs:
            LOG.debug('SPDK lvs name: %s, total space: %s, free space: %s',
                      lvs['name'],
                      lvs['total_size'],
                      lvs['free_size'])

    def _get_spdk_volume_name(self, name):
        output = self._rpc_call('bdev_get_bdevs')
        for bdev in output:
            for alias in bdev['aliases']:
                if name in alias:
                    return bdev['name']

    def _get_spdk_lvs_uuid(self, spdk_name):
        output = self._rpc_call('bdev_get_bdevs')
        for bdev in output:
            if spdk_name in bdev['name']:
                return bdev['driver_specific']['lvol']['lvol_store_uuid']

    def _get_spdk_lvs_free_space(self, lvs_uuid):
        self._update_volume_stats()

        for lvs in self.lvs:
            if lvs_uuid in lvs['uuid']:
                return lvs['free_size']

        return 0

    def _delete_bdev(self, name):
        spdk_name = self._get_spdk_volume_name(name)
        if spdk_name is not None:
            params = {'name': spdk_name}
            self._rpc_call('bdev_lvol_delete', params)
            LOG.debug('SPDK bdev %s deleted', spdk_name)
        else:
            LOG.debug('Could not find volume %s using SPDK driver', name)

    def _create_volume(self, volume, snapshot=None):
        output = self._rpc_call('bdev_lvol_get_lvstores')
        for lvs in output:
            free_size = (lvs['free_clusters'] * lvs['cluster_size'])
            if free_size / units.Gi >= volume.size:
                if snapshot is None:
                    params = {
                        'lvol_name': volume.name,
                        'size': volume.size * units.Gi,
                        'uuid': lvs['uuid']}
                    output2 = self._rpc_call('bdev_lvol_create', params)
                else:
                    snapshot_spdk_name = (
                        self._get_spdk_volume_name(snapshot.name))
                    params = {
                        'clone_name': volume.name,
                        'snapshot_name': snapshot_spdk_name}
                    output2 = self._rpc_call('bdev_lvol_clone', params)
                    spdk_name = self._get_spdk_volume_name(volume.name)
                    params = {'name': spdk_name}
                    self._rpc_call('bdev_lvol_inflate', params)

                    if volume.size > snapshot.volume_size:
                        params = {'name': spdk_name,
                                  'size': volume.size * units.Gi}
                        self._rpc_call('bdev_lvol_resize', params)

                LOG.debug('SPDK created lvol: %s', output2)

                return

        LOG.error('Unable to create volume using SPDK - no resources found')
        raise exception.VolumeBackendAPIException(
            data=_('Unable to create volume using SPDK'
                   ' - no resources found'))

    def do_setup(self, context):
        try:
            payload = {'method': 'bdev_get_bdevs', 'jsonrpc': '2.0', 'id': 1}
            self.url = ('%(protocol)s://%(ip)s:%(port)s/' %
                        {'protocol': self.configuration.spdk_rpc_protocol,
                         'ip': self.configuration.spdk_rpc_ip,
                         'port': self.configuration.spdk_rpc_port})
            requests.post(self.url,
                          data=json.dumps(payload),
                          auth=(self.configuration.spdk_rpc_username,
                                self.configuration.spdk_rpc_password),
                          verify=self.configuration.driver_ssl_cert_verify,
                          timeout=30)
        except Exception as err:
            err_msg = (
                _('Could not connect to SPDK target: %(err)s')
                % {'err': err})
            LOG.error(err_msg)
            raise exception.VolumeBackendAPIException(data=err_msg)

    def check_for_setup_error(self):
        """Verify that requirements are in place to use LVM driver."""

        # If configuration is incorrect we will get exception here
        self._rpc_call('bdev_get_bdevs')

    def create_volume(self, volume):
        """Creates a logical volume."""
        LOG.debug('SPDK create volume')

        return self._create_volume(volume)

    def delete_volume(self, volume):
        """Deletes a logical volume."""
        LOG.debug('SPDK deleting volume %s', volume.name)

        self._delete_bdev(volume.name)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot."""

        free_size = self._get_spdk_lvs_free_space(
            self._get_spdk_lvs_uuid(
                self._get_spdk_volume_name(snapshot.name)))

        if free_size < volume.size:
            raise exception.VolumeBackendAPIException(
                data=_('Not enough space to create snapshot with SPDK'))

        return self._create_volume(volume, snapshot)

    def create_snapshot(self, snapshot):
        """Creates a snapshot."""
        volume = snapshot['volume']
        spdk_name = self._get_spdk_volume_name(volume.name)

        if spdk_name is None:
            raise exception.VolumeBackendAPIException(
                data=_('Could not create snapshot with SPDK driver'))

        free_size = self._get_spdk_lvs_free_space(
            self._get_spdk_lvs_uuid(spdk_name))

        if free_size < volume.size:
            raise exception.VolumeBackendAPIException(
                data=_('Not enough space to create snapshot with SPDK'))

        params = {
            'lvol_name': spdk_name,
            'snapshot_name': snapshot['name']}
        self._rpc_call('bdev_lvol_snapshot', params)

        params = {'name': spdk_name}
        self._rpc_call('bdev_lvol_inflate', params)

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""
        spdk_name = self._get_spdk_volume_name(snapshot.name)
        if spdk_name is None:
            return

        params = {'name': spdk_name}
        bdev = self._rpc_call('bdev_get_bdevs', params)
        if 'clones' in bdev[0]['driver_specific']['lvol']:
            for clone in bdev[0]['driver_specific']['lvol']['clones']:
                spdk_name = self._get_spdk_volume_name(clone)
                params = {'name': spdk_name}
                self._rpc_call('bdev_lvol_inflate', params)

        self._delete_bdev(snapshot.name)

    def create_cloned_volume(self, volume, src_volume):
        spdk_name = self._get_spdk_volume_name(src_volume.name)

        free_size = self._get_spdk_lvs_free_space(
            self._get_spdk_lvs_uuid(spdk_name))

        # We need additional space for snapshot that will be used here
        if free_size < 2 * src_volume.size + volume.size:
            raise exception.VolumeBackendAPIException(
                data=_('Not enough space to clone volume with SPDK'))

        snapshot_name = 'snp-' + src_volume.name

        params = {
            'lvol_name': spdk_name,
            'snapshot_name': snapshot_name}
        self._rpc_call('bdev_lvol_snapshot', params)

        params = {'name': spdk_name}
        self._rpc_call('bdev_lvol_inflate', params)

        snapshot_spdk_name = self._get_spdk_volume_name(snapshot_name)
        params = {
            'clone_name': volume.name,
            'snapshot_name': snapshot_spdk_name}

        self._rpc_call('bdev_lvol_clone', params)

        spdk_name = self._get_spdk_volume_name(volume.name)
        params = {'name': spdk_name}
        self._rpc_call('bdev_lvol_inflate', params)

        self._delete_bdev(snapshot_name)

        if volume.size > src_volume.size:
            self.extend_volume(volume, volume.size)

    def copy_image_to_volume(self, context, volume, image_service, image_id):
        """Fetch the image from image_service and write it to the volume."""

        volume['provider_location'] = (
            self.create_export(context, volume, None)['provider_location'])
        connection_data = self.initialize_connection(volume, None)['data']
        target_connector = (
            connector.InitiatorConnector.factory(initiator.NVME,
                                                 utils.get_root_helper()))

        try:
            device_info = target_connector.connect_volume(connection_data)
        except Exception:
            LOG.info('Could not connect SPDK target device')
            return

        connection_data['device_path'] = device_info['path']

        try:
            image_utils.fetch_to_raw(context,
                                     image_service,
                                     image_id,
                                     device_info['path'],
                                     self.configuration.volume_dd_blocksize,
                                     size=volume['size'])

        finally:
            target_connector.disconnect_volume(connection_data, volume)

    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        """Copy the volume to the specified image."""
        volume['provider_location'] = (
            self.create_export(context, volume, None)['provider_location'])
        connection_data = self.initialize_connection(volume, None)['data']
        target_connector = (
            connector.InitiatorConnector.factory(initiator.NVME,
                                                 utils.get_root_helper()))

        try:
            device_info = target_connector.connect_volume(connection_data)
        except Exception:
            LOG.info('Could not connect SPDK target device')
            return

        connection_data['device_path'] = device_info['path']

        try:
            volume_utils.upload_volume(context,
                                       image_service,
                                       image_meta,
                                       device_info['path'],
                                       volume)
        finally:
            target_connector.disconnect_volume(connection_data, volume)

    def extend_volume(self, volume, new_size):
        """Extend an existing volume's size."""
        spdk_name = self._get_spdk_volume_name(volume.name)
        params = {'name': spdk_name, 'size': new_size * units.Gi}
        self._rpc_call('bdev_lvol_resize', params)

    # #######  Interface methods for DataPath (Target Driver) ########
    def ensure_export(self, context, volume):
        pass

    def create_export(self, context, volume, connector, vg=None):
        export_info = self.target_driver.create_export(
            context,
            volume,
            None)
        return {'provider_location': export_info['location'],
                'provider_auth': export_info['auth'], }

    def remove_export(self, context, volume):
        self.target_driver.remove_export(context, volume)

    def initialize_connection(self, volume, connector):
        return self.target_driver.initialize_connection(volume, connector)

    def validate_connector(self, connector):
        return self.target_driver.validate_connector(connector)

    def terminate_connection(self, volume, connector, **kwargs):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from six.moves import http_client

from cinder.api import extensions
from cinder.api import microversions as mv
from cinder.api.v3 import messages
from cinder import context
from cinder import exception
from cinder.message import api as message_api
from cinder.message import message_field
from cinder.tests.unit.api import fakes
from cinder.tests.unit.api.v3 import fakes as v3_fakes
from cinder.tests.unit import test


NS = '{http://docs.openstack.org/api/openstack-block-storage/3.0/content}'


@ddt.ddt
class MessageApiTest(test.TestCase):
    def setUp(self):
        super(MessageApiTest, self).setUp()
        self.ext_mgr = extensions.ExtensionManager()
        self.ext_mgr.extensions = {}
        self.controller = messages.MessagesController(self.ext_mgr)

        self.maxDiff = None
        self.ctxt = context.RequestContext('admin', 'fakeproject', True)

    def _expected_message_from_controller(self, id):
        message = v3_fakes.fake_message(id)
        links = [
            {'href': 'http://localhost/v3/fakeproject/messages/%s' % id,
             'rel': 'self'},
            {'href': 'http://localhost/fakeproject/messages/%s' % id,
             'rel': 'bookmark'},
        ]
        return {
            'message': {
                'id': message.get('id'),
                'user_message': "%s:%s" % (
                    message_field.translate_action(message.get('action_id')),
                    message_field.translate_detail(message.get('detail_id'))),
                'request_id': message.get('request_id'),
                'event_id': message.get('event_id'),
                'created_at': message.get('created_at'),
                'message_level': message.get('message_level'),
                'guaranteed_until': message.get('expires_at'),
                'links': links,
            }
        }

    def test_show(self):
        self.mock_object(message_api.API, 'get', v3_fakes.fake_message_get)

        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)
        req.environ['cinder.context'] = self.ctxt

        res_dict = self.controller.show(req, fakes.FAKE_UUID)

        ex = self._expected_message_from_controller(fakes.FAKE_UUID)
        self.assertEqual(ex, res_dict)

    def test_show_not_found(self):
        self.mock_object(message_api.API, 'get',
                         side_effect=exception.MessageNotFound(
                             message_id=fakes.FAKE_UUID))

        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)
        req.environ['cinder.context'] = self.ctxt

        self.assertRaises(exception.MessageNotFound, self.controller.show,
                          req, fakes.FAKE_UUID)

    def test_show_pre_microversion(self):
        self.mock_object(message_api.API, 'get', v3_fakes.fake_message_get)

        req = fakes.HTTPRequest.blank('/v3/messages/%s' % fakes.FAKE_UUID,
                                      version=mv.BASE_VERSION)
        req.environ['cinder.context'] = self.ctxt

        self.assertRaises(exception.VersionNotFoundForAPIMethod,
                          self.controller.show, req, fakes.FAKE_UUID)

    def test_delete(self):
        self.mock_object(message_api.API, 'get', v3_fakes.fake_message_get)
        self.mock_object(message_api.API, 'delete')

        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)
        req.environ['cinder.context'] = self.ctxt

        resp = self.controller.delete(req, fakes.FAKE_UUID)

        self.assertEqual(http_client.NO_CONTENT, resp.status_int)
        self.assertTrue(message_api.API.delete.called)

    def test_delete_not_found(self):
        self.mock_object(message_api.API, 'get',
                         side_effect=exception.MessageNotFound(
                             message_id=fakes.FAKE_UUID))

        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)

        self.assertRaises(exception.MessageNotFound, self.controller.delete,
                          req, fakes.FAKE_UUID)

    @ddt.data(mv.get_prior_version(mv.RESOURCE_FILTER),
              mv.RESOURCE_FILTER, mv.LIKE_FILTER)
    @mock.patch('cinder.api.common.reject_invalid_filters')
    def test_message_list_with_general_filter(self, version, mock_update):
        url = '/v3/%s/messages' % fakes.FAKE_UUID
        req = fakes.HTTPRequest.blank(url,
                                      version=version,
                                      use_admin_context=False)
        self.controller.index(req)

        if version != mv.get_prior_version(mv.RESOURCE_FILTER):
            support_like = True if version == mv.LIKE_FILTER else False
            mock_update.assert_called_once_with(req.environ['cinder.context'],
                                                mock.ANY, 'message',
                                                support_like)

    def test_index(self):
        self.mock_object(message_api.API, 'get_all',
                         return_value=[v3_fakes.fake_message(fakes.FAKE_UUID)])
        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)
        req.environ['cinder.context'] = self.ctxt

        res_dict = self.controller.index(req)

        ex = self._expected_message_from_controller(fakes.FAKE_UUID)
        expected = {
            'messages': [ex['message']]
        }
        self.assertDictEqual(expected, res_dict)



# Copyright (c) 2016 Clinton Knight
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume.drivers.netapp import options as na_opts

SSC_VSERVER = 'fake_vserver'
SSC_VOLUMES = ('volume1', 'volume2')
SSC_VOLUME_MAP = {
    SSC_VOLUMES[0]: {
        'pool_name': SSC_VOLUMES[0],
    },
    SSC_VOLUMES[1]: {
        'pool_name': SSC_VOLUMES[1],
    },
}
SSC_AGGREGATES = ('aggr1', 'aggr2')

SSC = {
    'volume1': {
        'thick_provisioning_support': True,
        'thin_provisioning_support': False,
        'netapp_thin_provisioned': 'false',
        'netapp_aggregate': 'aggr1',
        'netapp_compression': 'false',
        'netapp_dedup': 'true',
        'netapp_mirrored': 'false',
        'netapp_raid_type': 'raid_dp',
        'netapp_disk_type': ['SSD'],
        'netapp_hybrid_aggregate': 'false',
        'netapp_flexvol_encryption': 'true',
        'pool_name': 'volume1',
    },
    'volume2': {
        'thick_provisioning_support': False,
        'thin_provisioning_support': True,
        'netapp_thin_provisioned': 'true',
        'netapp_aggregate': 'aggr2',
        'netapp_compression': 'true',
        'netapp_dedup': 'true',
        'netapp_mirrored': 'true',
        'netapp_raid_type': 'raid_dp',
        'netapp_disk_type': ['FCAL', 'SSD'],
        'netapp_hybrid_aggregate': 'true',
        'netapp_flexvol_encryption': 'false',
        'pool_name': 'volume2',
    },
}

SSC_FLEXVOL_INFO = {
    'volume1': {
        'thick_provisioning_support': True,
        'thin_provisioning_support': False,
        'netapp_thin_provisioned': 'false',
        'netapp_aggregate': 'aggr1',
    },
    'volume2': {
        'thick_provisioning_support': False,
        'thin_provisioning_support': True,
        'netapp_thin_provisioned': 'true',
        'netapp_aggregate': 'aggr2',
    },
}

SSC_DEDUPE_INFO = {
    'volume1': {
        'netapp_dedup': 'true',
        'netapp_compression': 'false',
    },
    'volume2': {
        'netapp_dedup': 'true',
        'netapp_compression': 'true',
    },
}

SSC_ENCRYPTION_INFO = {
    'volume1': {
        'netapp_flexvol_encryption': 'true',
    },
    'volume2': {
        'netapp_flexvol_encryption': 'false',
    },
}

SSC_MIRROR_INFO = {
    'volume1': {
        'netapp_mirrored': 'false',
    },
    'volume2': {
        'netapp_mirrored': 'true',
    },
}

SSC_AGGREGATE_INFO = {
    'volume1': {
        'netapp_disk_type': ['SSD'],
        'netapp_raid_type': 'raid_dp',
        'netapp_hybrid_aggregate': 'false',
    },
    'volume2': {
        'netapp_disk_type': ['FCAL', 'SSD'],
        'netapp_raid_type': 'raid_dp',
        'netapp_hybrid_aggregate': 'true',
    },
}

PROVISIONING_OPTS = {
    'aggregate': 'fake_aggregate',
    'thin_provisioned': True,
    'snapshot_policy': None,
    'language': 'en_US',
    'dedupe_enabled': False,
    'compression_enabled': False,
    'snapshot_reserve': '12',
    'volume_type': 'rw',
    'size': 20,
}

ENCRYPTED_PROVISIONING_OPTS = {
    'aggregate': 'fake_aggregate',
    'thin_provisioned': True,
    'snapshot_policy': None,
    'language': 'en_US',
    'dedupe_enabled': False,
    'compression_enabled': False,
    'snapshot_reserve': '12',
    'volume_type': 'rw',
    'size': 20,
    'encrypt': 'true',
}


def get_fake_cmode_config(backend_name):

    config = configuration.Configuration(driver.volume_opts,
                                         config_group=backend_name)
    config.append_config_values(na_opts.netapp_proxy_opts)
    config.append_config_values(na_opts.netapp_connection_opts)
    config.append_config_values(na_opts.netapp_transport_opts)
    config.append_config_values(na_opts.netapp_basicauth_opts)
    config.append_config_values(na_opts.netapp_provisioning_opts)
    config.append_config_values(na_opts.netapp_cluster_opts)
    config.append_config_values(na_opts.netapp_san_opts)
    config.append_config_values(na_opts.netapp_replication_opts)
    config.append_config_values(na_opts.netapp_support_opts)

    return config



# Copyright (c) 2016 - 2019 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
from unittest import mock

import ddt

from cinder import exception
from cinder.volume import configuration as conf
from cinder.volume.drivers.dell_emc.unity import adapter as unity_adapter
from cinder.volume.drivers.dell_emc.unity import driver
from cinder.volume.drivers.dell_emc.unity import replication
from cinder.volume.drivers.san.san import san_opts


@ddt.ddt
class UnityReplicationTest(unittest.TestCase):
    @ddt.data({'version': '1.0.0', 'protocol': 'FC',
               'expected': unity_adapter.FCAdapter},
              {'version': '2.0.0', 'protocol': 'iSCSI',
               'expected': unity_adapter.ISCSIAdapter})
    @ddt.unpack
    def test_init_adapter(self, version, protocol, expected):
        a = replication.init_adapter(version, protocol)
        self.assertIsInstance(a, expected)
        self.assertEqual(version, a.version)


@ddt.ddt
class UnityReplicationDeviceTest(unittest.TestCase):
    def setUp(self):
        self.config = conf.Configuration(san_opts,
                                         config_group='unity-backend')
        self.config.san_ip = '1.1.1.1'
        self.config.san_login = 'user1'
        self.config.san_password = 'password1'
        self.driver = driver.UnityDriver(configuration=self.config)

        conf_dict = {'backend_id': 'secondary_unity', 'san_ip': '2.2.2.2'}
        self.mock_adapter = mock.MagicMock(is_setup=False)

        def mock_do_setup(*args):
            self.mock_adapter.is_setup = True

        self.mock_adapter.do_setup = mock.MagicMock(side_effect=mock_do_setup)
        with mock.patch('cinder.volume.drivers.dell_emc.unity.'
                        'replication.init_adapter',
                        return_value=self.mock_adapter):
            self.replication_device = replication.ReplicationDevice(
                conf_dict, self.driver)

    @ddt.data(
        {
            'conf_dict': {
                'backend_id': 'secondary_unity',
                'san_ip': '2.2.2.2'
            },
            'expected': [
                'secondary_unity', '2.2.2.2', 'user1', 'password1', 60
            ]
        },
        {
            'conf_dict': {
                'backend_id': 'secondary_unity',
                'san_ip': '2.2.2.2',
                'san_login': 'user2',
                'san_password': 'password2',
                'max_time_out_of_sync': 180
            },
            'expected': [
                'secondary_unity', '2.2.2.2', 'user2', 'password2', 180
            ]
        },
    )
    @ddt.unpack
    def test_init(self, conf_dict, expected):
        self.driver.configuration.replication_device = conf_dict
        device = replication.ReplicationDevice(conf_dict, self.driver)

        self.assertListEqual(
            [device.backend_id, device.san_ip, device.san_login,
             device.san_password, device.max_time_out_of_sync],
            expected)

        self.assertIs(self.driver, device.driver)

    @ddt.data(
        {
            'conf_dict': {'san_ip': '2.2.2.2'},
        },
        {
            'conf_dict': {'backend_id': '  ', 'san_ip': '2.2.2.2'},
        },
        {
            'conf_dict': {'backend_id': 'secondary_unity'},
        },
        {
            'conf_dict': {'backend_id': 'secondary_unity', 'san_ip': '  '},
        },
        {
            'conf_dict': {
                'backend_id': 'secondary_unity',
                'san_ip': '2.2.2.2',
                'san_login': 'user2',
                'san_password': 'password2',
                'max_time_out_of_sync': 'NOT_A_NUMBER'
            },
        },
    )
    @ddt.unpack
    def test_init_raise(self, conf_dict):
        self.driver.configuration.replication_device = conf_dict
        self.assertRaisesRegex(exception.InvalidConfigurationValue,
                               'Value .* is not valid for configuration '
                               'option "unity-backend.replication_device"',
                               replication.ReplicationDevice,
                               conf_dict, self.driver)

    @ddt.data(
        {
            'conf_dict': {
                'backend_id': 'secondary_unity',
                'san_ip': '2.2.2.2'
            },
            'expected': [
                '2.2.2.2', 'user1', 'password1'
            ]
        },
        {
            'conf_dict': {
                'backend_id': 'secondary_unity',
                'san_ip': '2.2.2.2',
                'san_login': 'user2',
                'san_password': 'password2',
                'max_time_out_of_sync': 180
            },
            'expected': [
                '2.2.2.2', 'user2', 'password2'
            ]
        },
    )
    @ddt.unpack
    def test_device_conf(self, conf_dict, expected):
        self.driver.configuration.replication_device = conf_dict
        device = replication.ReplicationDevice(conf_dict, self.driver)

        c = device.device_conf
        self.assertListEqual([c.san_ip, c.san_login, c.san_password],
                             expected)

    def test_setup_adapter(self):
        self.replication_device.setup_adapter()

        # Not call adapter.do_setup after initial setup done.
        self.replication_device.setup_adapter()

        self.mock_adapter.do_setup.assert_called_once()

    def test_setup_adapter_fail(self):
        def f(*args):
            raise exception.VolumeBackendAPIException('adapter setup failed')

        self.mock_adapter.do_setup = mock.MagicMock(side_effect=f)

        with self.assertRaises(exception.VolumeBackendAPIException):
            self.replication_device.setup_adapter()

    def test_adapter(self):
        self.assertIs(self.mock_adapter, self.replication_device.adapter)
        self.mock_adapter.do_setup.assert_called_once()

    def test_destination_pool(self):
        self.mock_adapter.storage_pools_map = {'pool-1': 'pool-1'}
        self.assertEqual('pool-1', self.replication_device.destination_pool)


@ddt.ddt
class UnityReplicationManagerTest(unittest.TestCase):
    def setUp(self):
        self.config = conf.Configuration(san_opts,
                                         config_group='unity-backend')
        self.config.san_ip = '1.1.1.1'
        self.config.san_login = 'user1'
        self.config.san_password = 'password1'
        self.config.replication_device = [
            {'backend_id': 'secondary_unity', 'san_ip': '2.2.2.2'}
        ]
        self.driver = driver.UnityDriver(configuration=self.config)

        self.replication_manager = replication.ReplicationManager()

    @mock.patch('cinder.volume.drivers.dell_emc.unity.'
                'replication.ReplicationDevice.setup_adapter')
    def test_do_setup(self, mock_setup_adapter):
        self.replication_manager.do_setup(self.driver)
        calls = [mock.call(), mock.call()]

        default_device = self.replication_manager.default_device
        self.assertEqual('1.1.1.1', default_device.san_ip)
        self.assertEqual('user1', default_device.san_login)
        self.assertEqual('password1', default_device.san_password)

        devices = self.replication_manager.replication_devices
        self.assertEqual(1, len(devices))
        self.assertIn('secondary_unity', devices)
        rep_device = devices['secondary_unity']
        self.assertEqual('2.2.2.2', rep_device.san_ip)
        self.assertEqual('user1', rep_device.san_login)
        self.assertEqual('password1', rep_device.san_password)

        self.assertTrue(self.replication_manager.is_replication_configured)

        self.assertTrue(
            self.replication_manager.active_backend_id is None
            or self.replication_manager.active_backend_id == 'default')

        self.assertFalse(self.replication_manager.is_service_failed_over)

        active_adapter = self.replication_manager.active_adapter
        calls.append(mock.call())
        self.assertIs(default_device.adapter, active_adapter)
        calls.append(mock.call())
        mock_setup_adapter.assert_has_calls(calls)

    @mock.patch('cinder.volume.drivers.dell_emc.unity.'
                'replication.ReplicationDevice.setup_adapter')
    def test_do_setup_replication_not_configured(self, mock_setup_adapter):
        self.driver.configuration.replication_device = None

        self.replication_manager.do_setup(self.driver)
        calls = [mock.call()]

        default_device = self.replication_manager.default_device
        self.assertEqual('1.1.1.1', default_device.san_ip)
        self.assertEqual('user1', default_device.san_login)
        self.assertEqual('password1', default_device.san_password)

        devices = self.replication_manager.replication_devices
        self.assertEqual(0, len(devices))

        self.assertFalse(self.replication_manager.is_replication_configured)

        self.assertTrue(
            self.replication_manager.active_backend_id is None
            or self.replication_manager.active_backend_id == 'default')

        self.assertFalse(self.replication_manager.is_service_failed_over)

        active_adapter = self.replication_manager.active_adapter
        calls.append(mock.call())
        self.assertIs(default_device.adapter, active_adapter)
        calls.append(mock.call())

        mock_setup_adapter.assert_has_calls(calls)

    @mock.patch('cinder.volume.drivers.dell_emc.unity.'
                'replication.ReplicationDevice.setup_adapter')
    def test_do_setup_failed_over(self, mock_setup_adapter):
        self.driver = driver.UnityDriver(configuration=self.config,
                                         active_backend_id='secondary_unity')

        self.replication_manager.do_setup(self.driver)
        calls = [mock.call()]

        default_device = self.replication_manager.default_device
        self.assertEqual('1.1.1.1', default_device.san_ip)
        self.assertEqual('user1', default_device.san_login)
        self.assertEqual('password1', default_device.san_password)

        devices = self.replication_manager.replication_devices
        self.assertEqual(1, len(devices))
        self.assertIn('secondary_unity', devices)
        rep_device = devices['secondary_unity']
        self.assertEqual('2.2.2.2', rep_device.san_ip)
        self.assertEqual('user1', rep_device.san_login)
        self.assertEqual('password1', rep_device.san_password)

        self.assertTrue(self.replication_manager.is_replication_configured)

        self.assertEqual('secondary_unity',
                         self.replication_manager.active_backend_id)

        self.assertTrue(self.replication_manager.is_service_failed_over)

        active_adapter = self.replication_manager.active_adapter
        calls.append(mock.call())
        self.assertIs(rep_device.adapter, active_adapter)
        calls.append(mock.call())

        mock_setup_adapter.assert_has_calls(calls)

    @ddt.data(
        {
            'rep_device': [{
                'backend_id': 'default', 'san_ip': '2.2.2.2'
            }]
        },
        {
            'rep_device': [{
                'backend_id': 'secondary_unity', 'san_ip': '2.2.2.2'
            }, {
                'backend_id': 'default', 'san_ip': '3.3.3.3'
            }]
        },
        {
            'rep_device': [{
                'backend_id': 'secondary_unity', 'san_ip': '2.2.2.2'
            }, {
                'backend_id': 'third_unity', 'san_ip': '3.3.3.3'
            }]
        },
    )
    @ddt.unpack
    @mock.patch('cinder.volume.drivers.dell_emc.unity.'
                'replication.ReplicationDevice.setup_adapter')
    def test_do_setup_raise_invalid_rep_device(self, mock_setup_adapter,
                                               rep_device):
        self.driver.configuration.replication_device = rep_device

        self.assertRaises(exception.InvalidConfigurationValue,
                          self.replication_manager.do_setup,
                          self.driver)

    @mock.patch('cinder.volume.drivers.dell_emc.unity.'
                'replication.ReplicationDevice.setup_adapter')
    def test_do_setup_raise_invalid_active_backend_id(self,
                                                      mock_setup_adapter):
        self.driver = driver.UnityDriver(configuration=self.config,
                                         active_backend_id='third_unity')

        self.assertRaises(exception.InvalidConfigurationValue,
                          self.replication_manager.do_setup,
                          self.driver)

    @mock.patch('cinder.volume.drivers.dell_emc.unity.'
                'replication.ReplicationDevice.setup_adapter')
    def test_failover_service(self, mock_setup_adapter):

        self.assertIsNone(self.replication_manager.active_backend_id)

        self.replication_manager.do_setup(self.driver)
        self.replication_manager.active_adapter

        self.assertEqual('default',
                         self.replication_manager.active_backend_id)

        self.replication_manager.failover_service('secondary_unity')
        self.assertEqual('secondary_unity',
                         self.replication_manager.active_backend_id)



#    (c) Copyright 2014 Cisco Systems Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""
Common constants used by Cisco FC Zone Driver.
"""
ACTIVE_ZONE_CONFIG = 'active_zone_config'
CFG_ZONESET = 'zoneset'
CFG_ZONE = 'zone'
CFG_ZONE_MEMBER = 'pwwn'
CFG_ZONES = 'zones'
ZONE_ADD = 'zoneadd'
ZONE_REMOVE = 'zoneremove'
"""
CLI Commands for FC zoning operations.
"""
GET_ACTIVE_ZONE_CFG = 'show zoneset active vsan '
FCNS_SHOW = 'show fcns database vsan '
GET_ZONE_STATUS = 'show zone status vsan '



# Copyright 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from six.moves import http_client
import webob

from cinder.api import common
from cinder.api.openstack import wsgi
from cinder.api.schemas import volume_metadata as metadata
from cinder.api import validation
from cinder import exception
from cinder.i18n import _
from cinder import volume


class Controller(wsgi.Controller):
    """The volume metadata API controller for the OpenStack API."""

    def __init__(self):
        self.volume_api = volume.API()
        super(Controller, self).__init__()

    def _get_metadata(self, context, volume_id):
        # The metadata is at the second position of the tuple returned
        # from _get_volume_and_metadata
        return self._get_volume_and_metadata(context, volume_id)[1]

    def _get_volume_and_metadata(self, context, volume_id):
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, volume_id)
        meta = self.volume_api.get_volume_metadata(context, volume)
        return (volume, meta)

    def index(self, req, volume_id):
        """Returns the list of metadata for a given volume."""
        context = req.environ['cinder.context']
        return {'metadata': self._get_metadata(context, volume_id)}

    @validation.schema(metadata.create)
    def create(self, req, volume_id, body):
        context = req.environ['cinder.context']
        metadata = body['metadata']

        new_metadata = self._update_volume_metadata(context, volume_id,
                                                    metadata, delete=False,
                                                    use_create=True)
        return {'metadata': new_metadata}

    @validation.schema(metadata.update)
    def update(self, req, volume_id, id, body):
        meta_item = body['meta']

        if id not in meta_item:
            expl = _('Request body and URI mismatch')
            raise webob.exc.HTTPBadRequest(explanation=expl)

        if len(meta_item) > 1:
            expl = _('Request body contains too many items')
            raise webob.exc.HTTPBadRequest(explanation=expl)

        context = req.environ['cinder.context']
        self._update_volume_metadata(context,
                                     volume_id,
                                     meta_item,
                                     delete=False)

        return {'meta': meta_item}

    @validation.schema(metadata.create)
    def update_all(self, req, volume_id, body):
        metadata = body['metadata']
        context = req.environ['cinder.context']

        new_metadata = self._update_volume_metadata(context,
                                                    volume_id,
                                                    metadata,
                                                    delete=True)

        return {'metadata': new_metadata}

    def _update_volume_metadata(self, context, volume_id, metadata,
                                delete=False, use_create=False):
        try:
            volume = self.volume_api.get(context, volume_id)
            if use_create:
                return self.volume_api.create_volume_metadata(context, volume,
                                                              metadata)
            else:
                return self.volume_api.update_volume_metadata(
                    context, volume, metadata, delete,
                    meta_type=common.METADATA_TYPES.user)
        # Not found exception will be handled at the wsgi level
        except (ValueError, AttributeError):
            msg = _("Malformed request body")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        except exception.InvalidVolumeMetadata as error:
            raise webob.exc.HTTPBadRequest(explanation=error.msg)

        except exception.InvalidVolumeMetadataSize as error:
            raise webob.exc.HTTPRequestEntityTooLarge(explanation=error.msg)

    def show(self, req, volume_id, id):
        """Return a single metadata item."""
        context = req.environ['cinder.context']
        data = self._get_metadata(context, volume_id)

        try:
            return {'meta': {id: data[id]}}
        except KeyError:
            raise exception.VolumeMetadataNotFound(volume_id=volume_id,
                                                   metadata_key=id)

    def delete(self, req, volume_id, id):
        """Deletes an existing metadata."""
        context = req.environ['cinder.context']

        volume, metadata = self._get_volume_and_metadata(context, volume_id)

        if id not in metadata:
            raise exception.VolumeMetadataNotFound(volume_id=volume_id,
                                                   metadata_key=id)

        # Not found exception will be handled at the wsgi level
        self.volume_api.delete_volume_metadata(
            context,
            volume,
            id,
            meta_type=common.METADATA_TYPES.user)
        return webob.Response(status_int=http_client.OK)


def create_resource():
    return wsgi.Resource(Controller())



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import excutils
import taskflow.engines
from taskflow.patterns import linear_flow

from cinder import exception
from cinder import flow_utils
from cinder.message import api as message_api
from cinder.message import message_field
from cinder import rpc
from cinder import utils
from cinder.volume.flows import common

LOG = logging.getLogger(__name__)

ACTION = 'volume:create'


class ExtractSchedulerSpecTask(flow_utils.CinderTask):
    """Extracts a spec object from a partial and/or incomplete request spec.

    Reversion strategy: N/A
    """

    default_provides = set(['request_spec'])

    def __init__(self, **kwargs):
        super(ExtractSchedulerSpecTask, self).__init__(addons=[ACTION],
                                                       **kwargs)

    def _populate_request_spec(self, volume, snapshot_id, image_id, backup_id):
        # Create the full request spec using the volume object.
        #
        # NOTE(dulek): At this point, a volume can be deleted before it gets
        # scheduled.  If a delete API call is made, the volume gets instantly
        # delete and scheduling will fail when it tries to update the DB entry
        # (with the host) in ScheduleCreateVolumeTask below.
        volume_type_id = volume.volume_type_id
        vol_type = volume.volume_type
        return {
            'volume_id': volume.id,
            'snapshot_id': snapshot_id,
            'image_id': image_id,
            'backup_id': backup_id,
            'volume_properties': {
                'size': utils.as_int(volume.size, quiet=False),
                'availability_zone': volume.availability_zone,
                'volume_type_id': volume_type_id,
            },
            'volume_type': list(dict(vol_type).items()),
        }

    def execute(self, context, request_spec, volume, snapshot_id,
                image_id, backup_id):
        # For RPC version < 1.2 backward compatibility
        if request_spec is None:
            request_spec = self._populate_request_spec(volume,
                                                       snapshot_id, image_id,
                                                       backup_id)
        return {
            'request_spec': request_spec,
        }


class ScheduleCreateVolumeTask(flow_utils.CinderTask):
    """Activates a scheduler driver and handles any subsequent failures.

    Notification strategy: on failure the scheduler rpc notifier will be
    activated and a notification will be emitted indicating what errored,
    the reason, and the request (and misc. other data) that caused the error
    to be triggered.

    Reversion strategy: N/A
    """
    FAILURE_TOPIC = "scheduler.create_volume"

    def __init__(self, driver_api, **kwargs):
        super(ScheduleCreateVolumeTask, self).__init__(addons=[ACTION],
                                                       **kwargs)
        self.driver_api = driver_api
        self.message_api = message_api.API()

    def _handle_failure(self, context, request_spec, cause):
        try:
            self._notify_failure(context, request_spec, cause)
        finally:
            LOG.error("Failed to run task %(name)s: %(cause)s",
                      {'cause': cause, 'name': self.name})

    @utils.if_notifications_enabled
    def _notify_failure(self, context, request_spec, cause):
        """When scheduling fails send out an event that it failed."""
        payload = {
            'request_spec': request_spec,
            'volume_properties': request_spec.get('volume_properties', {}),
            'volume_id': request_spec['volume_id'],
            'state': 'error',
            'method': 'create_volume',
            'reason': cause,
        }
        try:
            rpc.get_notifier('scheduler').error(context, self.FAILURE_TOPIC,
                                                payload)
        except exception.CinderException:
            LOG.exception("Failed notifying on %(topic)s "
                          "payload %(payload)s",
                          {'topic': self.FAILURE_TOPIC, 'payload': payload})

    def execute(self, context, request_spec, filter_properties, volume):
        try:
            self.driver_api.schedule_create_volume(context, request_spec,
                                                   filter_properties)
        except Exception as e:
            self.message_api.create(
                context,
                message_field.Action.SCHEDULE_ALLOCATE_VOLUME,
                resource_uuid=request_spec['volume_id'],
                exception=e)
            # An error happened, notify on the scheduler queue and log that
            # this happened and set the volume to errored out and reraise the
            # error *if* exception caught isn't NoValidBackend. Otherwise *do
            # not* reraise (since what's the point?)
            with excutils.save_and_reraise_exception(
                    reraise=not isinstance(e, exception.NoValidBackend)):
                try:
                    self._handle_failure(context, request_spec, e)
                finally:
                    common.error_out(volume, reason=e)


def get_flow(context, driver_api, request_spec=None,
             filter_properties=None,
             volume=None, snapshot_id=None, image_id=None, backup_id=None):

    """Constructs and returns the scheduler entrypoint flow.

    This flow will do the following:

    1. Inject keys & values for dependent tasks.
    2. Extract a scheduler specification from the provided inputs.
    3. Use provided scheduler driver to select host and pass volume creation
       request further.
    """
    create_what = {
        'context': context,
        'raw_request_spec': request_spec,
        'filter_properties': filter_properties,
        'volume': volume,
        'snapshot_id': snapshot_id,
        'image_id': image_id,
        'backup_id': backup_id,
    }

    flow_name = ACTION.replace(":", "_") + "_scheduler"
    scheduler_flow = linear_flow.Flow(flow_name)

    # This will extract and clean the spec from the starting values.
    scheduler_flow.add(ExtractSchedulerSpecTask(
        rebind={'request_spec': 'raw_request_spec'}))

    # This will activate the desired scheduler driver (and handle any
    # driver related failures appropriately).
    scheduler_flow.add(ScheduleCreateVolumeTask(driver_api))

    # Now load (but do not run) the flow using the provided initial data.
    return taskflow.engines.load(scheduler_flow, store=create_what)



# Copyright 2016 Dell Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Core backup driver interface.

All backup drivers should support this interface as a bare minimum.
"""

from cinder.interface import base


class BackupDriver(base.CinderInterface):
    """Backup driver required interface."""

    def get_metadata(self, volume_id):
        """Get volume metadata.

        Returns a json-encoded dict containing all metadata and the restore
        version i.e. the version used to decide what actually gets restored
        from this container when doing a backup restore.

        Typically best to use py:class:`BackupMetadataAPI` for this.

        :param volume_id: The ID of the volume.
        :returns: json-encoded dict of metadata.
        """

    def put_metadata(self, volume_id, json_metadata):
        """Set volume metadata.

        Typically best to use py:class:`BackupMetadataAPI` for this.

        :param volume_id: The ID of the volume.
        :param json_metadata: The json-encoded dict of metadata.
        """

    def backup(self, backup, volume_file, backup_metadata=False):
        """Start a backup of a specified volume.

        If backup['parent_id'] is given, then an incremental backup
        should be performed.

        If the parent backup is of different size, a full backup should be
        performed to ensure all data is included.

        :param backup: The backup information.
        :param volume_file: The volume or file to write the backup to.
        :param backup_metadata: Whether to include volume metadata in the
                                backup.

        The variable structure of backup in the following format::

           {
              'id': id,
              'availability_zone': availability_zone,
              'service': driver_name,
              'user_id': context.user_id,
              'project_id': context.project_id,
              'display_name': name,
              'display_description': description,
              'volume_id': volume_id,
              'status': fields.BackupStatus.CREATING,
              'container': container,
              'parent_id': parent_id,
              'size': size,
              'host': host,
              'snapshot_id': snapshot_id,
              'data_timestamp': data_timestamp,
           }

        service: backup driver
        parent_id: parent backup id
        size: equal to volume size
        data_timestamp: backup creation time
        """

    def restore(self, backup, volume_id, volume_file):
        """Restore volume from a backup.

        :param backup: The backup information.
        :param volume_id: The volume to be restored.
        :param volume_file: The volume or file to read the data from.
        """

    def delete_backup(self, backup):
        """Delete a backup from the backup store.

        :param backup: The backup to be deleted.
        """

    def export_record(self, backup):
        """Export driver specific backup record information.

        If backup backend needs additional driver specific information to
        import backup record back into the system it must override this method
        and return it as a dictionary so it can be serialized into a string.

        Default backup driver implementation has no extra information.

        :param backup: backup object to export
        :returns: driver_info - dictionary with extra information
        """

    def import_record(self, backup, driver_info):
        """Import driver specific backup record information.

        If backup backend needs additional driver specific information to
        import backup record back into the system it must override this method
        since it will be called with the extra information that was provided by
        export_record when exporting the backup.

        Default backup driver implementation does nothing since it didn't
        export any specific data in export_record.

        :param backup: backup object to export
        :param driver_info: dictionary with driver specific backup record
                            information
        :returns: None
        """

    def check_for_setup_error(self):
        """Method for checking if backup backend is successfully installed.

        Depends on storage backend limitations and driver implementation this
        method could check if all needed config options are configurated well
        or try to connect to the storage to verify driver can do it without
        any issues.

        :returns: None
        """



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import base64
import binascii
from distutils import version
import math

from oslo_log import log as logging
from oslo_utils import units

LOG = logging.getLogger(__name__)


def version_gte(ver1, ver2):
    return version.LooseVersion(ver1) >= version.LooseVersion(ver2)


def convert_kb_to_gib(size):
    return int(math.floor(float(size) / units.Mi))


def id_to_base64(_id):
    # Base64 encode the id to get a volume name less than 32 characters due
    # to PowerFlex limitation.
    name = str(_id).replace("-", "")
    try:
        name = base64.b16decode(name.upper())
    except (TypeError, binascii.Error):
        pass
    if isinstance(name, str):
        name = name.encode()
    encoded_name = base64.b64encode(name).decode()
    LOG.debug("Converted id %(id)s to PowerFlex OS name %(name)s.",
              {"id": _id, "name": encoded_name})
    return encoded_name


def round_to_num_gran(size, num=8):
    """Round size to nearest value that is multiple of `num`."""

    if size % num == 0:
        return size
    return size + num - (size % num)


def round_down_to_num_gran(size, num=8):
    """Round size down to nearest value that is multiple of `num`."""

    return size - (size % num)



# Copyright (c) 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from oslo_log import log as logging

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.huawei import constants


LOG = logging.getLogger(__name__)


class FCZoneHelper(object):
    """FC zone helper for Huawei driver."""

    def __init__(self, fcsan_lookup_service, client):
        self.fcsan = fcsan_lookup_service
        self.client = client

    def _get_fc_ports_info(self):
        ports_info = {}
        data = self.client.get_fc_ports_on_array()
        for item in data:
            if item['RUNNINGSTATUS'] == constants.FC_PORT_CONNECTED:
                location = item['PARENTID'].split('.')
                port_info = {}
                port_info['id'] = item['ID']
                port_info['contr'] = location[0]
                port_info['bandwidth'] = item['RUNSPEED']
                ports_info[item['WWN']] = port_info
        return ports_info

    def _count_port_weight(self, port, ports_info):
        LOG.debug("Count weight for port: %s.", port)
        portgs = self.client.get_portgs_by_portid(ports_info[port]['id'])
        LOG.debug("Port %(port)s belongs to PortGroup %(portgs)s.",
                  {"port": port, "portgs": portgs})
        weight = 0
        for portg in portgs:
            views = self.client.get_views_by_portg(portg)
            if not views:
                LOG.debug("PortGroup %s doesn't belong to any view.", portg)
                continue

            LOG.debug("PortGroup %(portg)s belongs to view %(views)s.",
                      {"portg": portg, "views": views[0]})
            # In fact, there is just one view for one port group.
            lungroup = self.client.get_lungroup_by_view(views[0])
            lun_num = self.client.get_obj_count_from_lungroup(lungroup)
            ports_in_portg = self.client.get_ports_by_portg(portg)
            LOG.debug("PortGroup %(portg)s contains ports: %(ports)s.",
                      {"portg": portg, "ports": ports_in_portg})
            total_bandwidth = 0
            for port_pg in ports_in_portg:
                if port_pg in ports_info:
                    total_bandwidth += int(ports_info[port_pg]['bandwidth'])

            LOG.debug("Total bandwidth for PortGroup %(portg)s is %(bindw)s.",
                      {"portg": portg, "bindw": total_bandwidth})

            if total_bandwidth:
                weight += float(lun_num) / float(total_bandwidth)

        bandwidth = float(ports_info[port]['bandwidth'])
        return (weight, 10000 / bandwidth)

    def _get_weighted_ports_per_contr(self, ports, ports_info):
        port_weight_map = {}
        for port in ports:
            port_weight_map[port] = self._count_port_weight(port, ports_info)

        LOG.debug("port_weight_map: %s", port_weight_map)
        sorted_ports = sorted(port_weight_map.items(), key=lambda d: d[1])
        weighted_ports = []
        count = 0
        for port in sorted_ports:
            if count >= constants.PORT_NUM_PER_CONTR:
                break
            weighted_ports.append(port[0])
            count += 1
        return weighted_ports

    def _get_weighted_ports(self, contr_port_map, ports_info, contrs):
        LOG.debug("_get_weighted_ports, we only select ports from "
                  "controllers: %s", contrs)
        weighted_ports = []
        for contr in contrs:
            if contr in contr_port_map:
                weighted_ports_per_contr = self._get_weighted_ports_per_contr(
                    contr_port_map[contr], ports_info)
                LOG.debug("Selected ports %(ports)s on controller %(contr)s.",
                          {"ports": weighted_ports_per_contr,
                           "contr": contr})
                weighted_ports.extend(weighted_ports_per_contr)
        return weighted_ports

    def _filter_by_fabric(self, wwns, ports):
        """Filter FC ports and initiators connected to fabrics."""
        ini_tgt_map = self.fcsan.get_device_mapping_from_network(wwns, ports)
        fabric_connected_ports = []
        fabric_connected_initiators = []
        for fabric in ini_tgt_map:
            fabric_connected_ports.extend(
                ini_tgt_map[fabric]['target_port_wwn_list'])
            fabric_connected_initiators.extend(
                ini_tgt_map[fabric]['initiator_port_wwn_list'])

        if not fabric_connected_ports:
            msg = _("No FC port connected to fabric.")
            raise exception.VolumeBackendAPIException(data=msg)
        if not fabric_connected_initiators:
            msg = _("No initiator connected to fabric.")
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug("Fabric connected ports: %(ports)s, "
                  "Fabric connected initiators: %(initiators)s.",
                  {'ports': fabric_connected_ports,
                   'initiators': fabric_connected_initiators})
        return fabric_connected_ports, fabric_connected_initiators

    def _get_lun_engine_contrs(self, engines, lun_id,
                               lun_type=constants.LUN_TYPE):
        contrs = []
        engine_id = None
        lun_info = self.client.get_lun_info(lun_id, lun_type)
        lun_contr_id = lun_info['OWNINGCONTROLLER']
        for engine in engines:
            contrs = json.loads(engine['NODELIST'])
            engine_id = engine['ID']
            if lun_contr_id in contrs:
                break

        LOG.debug("LUN %(lun_id)s belongs to engine %(engine_id)s. Engine "
                  "%(engine_id)s has controllers: %(contrs)s.",
                  {"lun_id": lun_id, "engine_id": engine_id, "contrs": contrs})
        return contrs, engine_id

    def _build_contr_port_map(self, fabric_connected_ports, ports_info):
        contr_port_map = {}
        for port in fabric_connected_ports:
            contr = ports_info[port]['contr']
            if not contr_port_map.get(contr):
                contr_port_map[contr] = []
            contr_port_map[contr].append(port)
        LOG.debug("Controller port map: %s.", contr_port_map)
        return contr_port_map

    def _create_new_portg(self, portg_name, engine_id):
        portg_id = self.client.get_tgt_port_group(portg_name)
        if portg_id:
            LOG.debug("Found port group %s not belonged to any view, "
                      "deleting it.", portg_name)
            ports = self.client.get_fc_ports_by_portgroup(portg_id)
            for port_id in ports.values():
                self.client.remove_port_from_portgroup(portg_id, port_id)
            self.client.delete_portgroup(portg_id)
        description = constants.PORTGROUP_DESCRIP_PREFIX + engine_id
        new_portg_id = self.client.create_portg(portg_name, description)
        return new_portg_id

    def build_ini_targ_map(self, wwns, host_id, lun_id,
                           lun_type=constants.LUN_TYPE):
        engines = self.client.get_all_engines()
        LOG.debug("Get array engines: %s", engines)

        contrs, engine_id = self._get_lun_engine_contrs(engines, lun_id,
                                                        lun_type)

        # Check if there is already a port group in the view.
        # If yes and have already considered the engine,
        # we won't change anything about the port group and zone.
        view_name = constants.MAPPING_VIEW_PREFIX + host_id
        portg_name = constants.PORTGROUP_PREFIX + host_id
        view_id = self.client.find_mapping_view(view_name)
        portg_info = self.client.get_portgroup_by_view(view_id)
        portg_id = portg_info[0]['ID'] if portg_info else None

        init_targ_map = {}
        if portg_id:
            description = portg_info[0].get("DESCRIPTION", '')
            engines = description.replace(constants.PORTGROUP_DESCRIP_PREFIX,
                                          "")
            engines = engines.split(',')
            ports = self.client.get_fc_ports_by_portgroup(portg_id)
            if engine_id in engines:
                LOG.debug("Have already selected ports for engine %s, just "
                          "use them.", engine_id)
                return (list(ports.keys()), portg_id, init_targ_map)

        # Filter initiators and ports that connected to fabrics.
        ports_info = self._get_fc_ports_info()
        (fabric_connected_ports, fabric_connected_initiators) = (
            self._filter_by_fabric(wwns, ports_info.keys()))

        # Build a controller->ports map for convenience.
        contr_port_map = self._build_contr_port_map(fabric_connected_ports,
                                                    ports_info)
        # Get the 'best' ports for the given controllers.
        weighted_ports = self._get_weighted_ports(contr_port_map, ports_info,
                                                  contrs)
        if not weighted_ports:
            msg = _("No FC port can be used for LUN %s.") % lun_id
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Handle port group.
        port_list = [ports_info[port]['id'] for port in weighted_ports]

        if portg_id:
            # Add engine ID to the description of the port group.
            self.client.append_portg_desc(portg_id, engine_id)
            # Extend the weighted_ports to include the ports already in the
            # port group.
            weighted_ports.extend(list(ports.keys()))
        else:
            portg_id = self._create_new_portg(portg_name, engine_id)

        for port in port_list:
            self.client.add_port_to_portg(portg_id, port)

        for ini in fabric_connected_initiators:
            init_targ_map[ini] = weighted_ports
        LOG.debug("build_ini_targ_map: Port group name: %(portg_name)s, "
                  "init_targ_map: %(map)s.",
                  {"portg_name": portg_name,
                   "map": init_targ_map})
        return weighted_ports, portg_id, init_targ_map

    def get_init_targ_map(self, wwns, host_id):
        error_ret = ([], None, {})
        if not host_id:
            return error_ret

        view_name = constants.MAPPING_VIEW_PREFIX + host_id
        view_id = self.client.find_mapping_view(view_name)
        if not view_id:
            return error_ret
        port_group = self.client.get_portgroup_by_view(view_id)
        portg_id = port_group[0]['ID'] if port_group else None
        ports = self.client.get_fc_ports_by_portgroup(portg_id)
        for port_id in ports.values():
            self.client.remove_port_from_portgroup(portg_id, port_id)
        init_targ_map = {}
        for wwn in wwns:
            init_targ_map[wwn] = list(ports.keys())
        return list(ports.keys()), portg_id, init_targ_map



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from six.moves import http_client

from cinder.api import microversions as mv
from cinder.tests.unit import fake_constants
from cinder.tests.unit.policies import test_base
from cinder.volume import api as volume_api


# TODO(yikun): The below policy test cases should be added:
# * REVERT_POLICY
# * RESET_STATUS
# * FORCE_DETACH_POLICY
# * UPLOAD_PUBLIC_POLICY
# * UPLOAD_IMAGE_POLICY
# * MIGRATE_POLICY
# * MIGRATE_COMPLETE_POLICY
class VolumeProtectionTests(test_base.CinderPolicyTests):
    def test_admin_can_extend_volume(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    def test_owner_can_extend_volume(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_extend_volume_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    def test_admin_can_extend_attached_volume(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(
            admin_context, path, 'POST', body=body,
            microversion=mv.VOLUME_EXTEND_INUSE)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    def test_owner_can_extend_attached_volume(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(
            user_context, path, 'POST', body=body,
            microversion=mv.VOLUME_EXTEND_INUSE)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_extend_attached_volume_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(
            non_owner_context, path, 'POST', body=body,
            microversion=mv.VOLUME_EXTEND_INUSE)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    def test_admin_can_retype_volume(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        vol_type = self._create_fake_type(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-retype": {"new_type": "%s" % vol_type.name,
                              "migration_policy": "never"}}
        response = self._get_request_response(
            admin_context, path, 'POST', body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    def test_owner_can_retype_volume(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        vol_type = self._create_fake_type(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-retype": {"new_type": "%s" % vol_type.name,
                              "migration_policy": "never"}}
        response = self._get_request_response(
            user_context, path, 'POST', body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_retype_volume_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        vol_type = self._create_fake_type(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-retype": {"new_type": "%s" % vol_type.name,
                              "migration_policy": "never"}}
        response = self._get_request_response(
            non_owner_context, path, 'POST', body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    def test_admin_can_update_readonly(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(
            admin_context, admin_metadata={"readonly": "False"})

        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-update_readonly_flag": {"readonly": "True"}}
        response = self._get_request_response(
            admin_context, path, 'POST', body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    def test_owner_can_update_readonly(self):
        user_context = self.user_context

        volume = self._create_fake_volume(
            user_context, admin_metadata={"readonly": "False"})

        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-update_readonly_flag": {"readonly": "True"}}
        response = self._get_request_response(
            user_context, path, 'POST', body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_update_readonly_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(
            user_context, admin_metadata={"readonly": "False"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-update_readonly_flag": {"readonly": "True"}}
        response = self._get_request_response(
            non_owner_context, path, 'POST', body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.API, 'get_volume')
    def test_admin_can_force_delete_volumes(self, mock_volume):
        # Make sure administrators are authorized to force delete volumes
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }
        body = {"os-force_delete": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)

        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get_volume')
    def test_nonadmin_cannot_force_delete_volumes(self, mock_volume):
        # Make sure volumes only can be force deleted by admin
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }
        body = {"os-force_delete": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)

        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'attach_volume')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'detach_volume')
    def test_admin_can_attach_detach_volume(self, mock_detach, mock_attach):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-attach": {"instance_uuid": fake_constants.UUID1,
                              "mountpoint": "/dev/vdc"}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

        body = {"os-detach": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'attach_volume')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'detach_volume')
    def test_owner_can_attach_detach_volume(self, mock_detach, mock_attach):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-attach": {"instance_uuid": fake_constants.UUID1,
                              "mountpoint": "/dev/vdc"}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

        body = {"os-detach": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'attach_volume')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'detach_volume')
    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_attach_detach_volume_for_others(self, mock_volume,
                                                          mock_detach,
                                                          mock_attach):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-attach": {"instance_uuid": fake_constants.UUID1,
                              "mountpoint": "/dev/vdc"}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

        body = {"os-detach": {}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    def test_admin_can_reserve_unreserve_volume(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-reserve": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

        body = {"os-unreserve": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    def test_owner_can_reserve_unreserve_volume(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-reserve": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

        body = {"os-unreserve": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_reserve_unreserve_volume_for_others(self,
                                                              mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-attach": {"instance_uuid": fake_constants.UUID1,
                              "mountpoint": "/dev/vdc"}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

        body = {"os-detach": {}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'initialize_connection')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'terminate_connection')
    def test_admin_can_initialize_terminate_conn(self, mock_t, mock_i):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-initialize_connection": {'connector': {}}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.OK, response.status_int)

        body = {"os-terminate_connection": {'connector': {}}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'initialize_connection')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'terminate_connection')
    def test_owner_can_initialize_terminate_conn(self, mock_t, mock_i):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-initialize_connection": {'connector': {}}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.OK, response.status_int)

        body = {"os-terminate_connection": {'connector': {}}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'initialize_connection')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'terminate_connection')
    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_initialize_terminate_conn_for_others(self,
                                                               mock_volume,
                                                               mock_t,
                                                               mock_i):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-initialize_connection": {'connector': {}}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

        body = {"os-terminate_connection": {'connector': {}}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    def test_admin_can_begin_roll_detaching(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, status='in-use',
                                          attach_status='attached')
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-begin_detaching": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

        body = {"os-roll_detaching": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    def test_owner_can_begin_roll_detaching(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, status='in-use',
                                          attach_status='attached')
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-begin_detaching": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

        body = {"os-roll_detaching": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_begin_roll_detaching_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context, status='in-use',
                                          attach_status='attached')
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-begin_detaching": {}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)

        body = {"os-roll_detaching": {}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(http_client.FORBIDDEN, response.status_int)



# Copyright (c) 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
:mod:`cinder.volume.drivers.san` -- Cinder San Drivers
=====================================================

.. automodule:: cinder.volume.drivers.san
   :platform: Unix
   :synopsis: Module containing all the Cinder San drivers.
"""

# Adding imports for backwards compatibility in loading volume_driver.
from cinder.volume.drivers.san.san import SanISCSIDriver      # noqa
